# -*- coding: utf-8 -*-
'''
Gaussian Mixture Model
----------------------
Fit a Gaussian Mixture Model (GMM) to data and draw samples from it. Uses the
Expectation-Maximization algorithm.
'''


from six.moves import range

import numpy as np
from scipy.stats import multivariate_normal
#from scipy.misc import logsumexp
from scipy.special import logsumexp
from . import multivariate_truncnorm as truncnorm
import itertools


# Equation references are from Numerical Recipes for general GMM and
# https://www.cs.nmsu.edu/~joemsong/publications/Song-SPIE2005-updated.pdf for
# online updating features


class estimator:
    '''
    Base estimator class for GMM

    Parameters
    ----------
    k : int
        Number of Gaussian components
    max_iters : int
        Maximum number of Expectation-Maximization iterations
    '''

    def __init__(self, k, max_iters=100, tempering_coeff=0.001):
        self.k = k # number of gaussian components
        self.max_iters = max_iters # maximum number of iterations to convergence
        self.means = [None] * k
        self.covariances =[None] * k
        self.weights = [None] * k
        self.d = None
        self.p_nk = None
        self.log_prob = None
        self.cov_avg_ratio = 0.05
        self.epsilon = 1e-4
        self.tempering_coeff = tempering_coeff

    def _initialize(self, n, sample_array, sample_weights=None):
        p_weights = (sample_weights / np.sum(sample_weights)).flatten()
        self.means = sample_array[np.random.choice(n, self.k, p=p_weights.astype(sample_array.dtype)), :]
        self.covariances = [np.identity(self.d)] * self.k
        self.weights = np.array([1.0 / self.k] * self.k)

    def _e_step(self, n, sample_array, sample_weights):
        '''
        Expectation step
        '''
        if sample_weights is None:
            log_sample_weights = np.zeros((n, 1))
        else:
            log_sample_weights = np.log(sample_weights)
        p_nk = np.empty((n, self.k))
        for index in range(self.k):
            mean = self.means[index]
            cov = self.covariances[index]
            log_p = np.log(self.weights[index])
            log_pdf = np.rot90([multivariate_normal.logpdf(x=sample_array, mean=mean, cov=cov, allow_singular=True)], -1) # (16.1.4)
            # note that allow_singular=True in the above line is probably really dumb and
            # terrible, but it seems to occasionally keep the whole thing from blowing up
            # so it stays for now
            p_nk[:,[index]] = log_pdf + log_p # (16.1.5)
        p_xn = logsumexp(p_nk, axis=1, keepdims=True) # (16.1.3)
        self.p_nk = p_nk - p_xn # (16.1.5)
        self.p_nk += log_sample_weights
        self.log_prob = np.sum(p_xn + log_sample_weights) # (16.1.2)

    def _m_step(self, n, sample_array):
        '''
        Maximization step
        '''
        p_nk = np.exp(self.p_nk)
        weights = np.sum(p_nk, axis=0)
        for index in range(self.k):
            # (16.1.6)
            w = weights[index]
            p_k = p_nk[:,[index]]
            mean = np.sum(np.multiply(sample_array, p_k), axis=0)
            mean /= w
            self.means[index] = mean
            # (16.1.6)
            diff = sample_array - mean
            cov = np.dot((p_k * diff).T, diff) / w
            # attempt to fix non-positive-semidefinite covariances
            self.covariances[index] = self._near_psd(cov)
            # (16.17)
        weights /= np.sum(p_nk)
        self.weights = weights


    def _tol(self, n):
        '''
        Scale tolerance with number of dimensions, number of components, and
        number of samples
        '''
        return (self.d * self.k * n) * 10e-4

    def _near_psd(self, x):
        '''
        Calculates the nearest postive semi-definite matrix for a correlation/covariance matrix
        
        Code from here:
        https://stackoverflow.com/questions/10939213/how-can-i-calculate-the-nearest-positive-semi-definite-matrix
        '''
        n = x.shape[0]
        var_list = np.array([np.sqrt(x[i,i]) for i in range(n)])
        y = np.array([[x[i, j]/(var_list[i]*var_list[j]) for i in range(n)] for j in range(n)])
        while True:
            epsilon = self.epsilon
            if min(np.linalg.eigvals(y)) > epsilon:
                return x

            # Removing scaling factor of covariance matrix
            var_list = np.array([np.sqrt(x[i,i]) for i in range(n)])
            y = np.array([[x[i, j]/(var_list[i]*var_list[j]) for i in range(n)] for j in range(n)])

            # getting the nearest correlation matrix
            eigval, eigvec = np.linalg.eig(y)
            val = np.matrix(np.maximum(eigval, epsilon))
            vec = np.matrix(eigvec)
            T = 1/(np.multiply(vec, vec) * val.T)
            T = np.matrix(np.sqrt(np.diag(np.array(T).reshape((n)) )))
            B = T * vec * np.diag(np.array(np.sqrt(val)).reshape((n)))
            near_corr = B*B.T    

            # returning the scaling factors
            near_cov = np.array([[near_corr[i, j]*(var_list[i]*var_list[j]) for i in range(n)] for j in range(n)])
            if np.isreal(near_cov).all():
                break
            else:
                x = near_cov.real
        return near_cov
    
    def fit(self, sample_array, sample_weights):
        '''
        Fit the model to data

        Parameters
        ----------
        sample_array : np.ndarray
            Array of samples to fit
        sample_weights : np.ndarray
            Weights for samples
        '''
        n, self.d = sample_array.shape
        self._initialize(n, sample_array, sample_weights)
        prev_log_prob = 0
        self.log_prob = float('inf')
        count = 0
        while abs(self.log_prob - prev_log_prob) > self._tol(n) and count < self.max_iters:
            prev_log_prob = self.log_prob
            self._e_step(n, sample_array, sample_weights)
            self._m_step(n, sample_array)
            count += 1
        for index in range(self.k):
            cov = self.covariances[index]
            # temper
            cov = (cov + self.tempering_coeff * np.eye(self.d)) / (1 + self.tempering_coeff)
            self.covariances[index] = cov

    def print_params(self):
        '''
        Prints the model's parameters in an easily-readable format
        '''
        for i in range(self.k):
            mean = self.means[i]
            cov = self.covariances[i]
            weight = self.weights[i]
            print('________________________________________\n')
            print('Component', i)
            print('Mean')
            print(mean)
            print('Covaraince')
            print(cov)
            print('Weight')
            print(weight, '\n')


class gmm:
    '''
    More sophisticated implementation built on top of estimator class

    Includes functionality to update with new data rather than re-fit, as well
    as sampling and scoring of samples.

    Parameters
    ----------
    k : int
        Number of Gaussian components
    max_iters : int
        Maximum number of Expectation-Maximization iterations
    '''

    def __init__(self, k, max_iters=1000):
        self.k = k
        #self.tol = tol
        self.max_iters = max_iters
        self.means = [None] * k
        self.covariances =[None] * k
        self.weights = [None] * k
        self.d = None
        self.p_nk = None
        self.log_prob = None
        self.N = 0
        self.epsilon = 1e-4  # allow very strong correlations
        self.tempering_coeff = 0.01

    def fit(self, sample_array, sample_weights=None):
        '''
        Fit the model to data

        Parameters
        ----------
        sample_array : np.ndarray
            Array of samples to fit
        sample_weights : np.ndarray
            Weights for samples
        '''
        self.N, self.d = sample_array.shape
        if sample_weights is None:
            sample_weights = np.ones((self.N, 1))
        # just use base estimator
        model = estimator(self.k, tempering_coeff=self.tempering_coeff)
        model.fit(sample_array, sample_weights)
        self.means = model.means
        self.covariances = model.covariances
        self.weights = model.weights
        self.p_nk = model.p_nk
        self.log_prob = model.log_prob

    def _match_components(self, new_model):
        '''
        Match components in new model to those in current model by minimizing the
        net Mahalanobis between all pairs of components
        '''
        orders = list(itertools.permutations(list(range(self.k)), self.k))
        distances = np.empty(len(orders))
        index = 0
        for order in orders:
            dist = 0
            i = 0
            for j in order:
                # get Mahalanobis distance between current pair of components
                diff = new_model.means[j] - self.means[i]
                cov_inv = np.linalg.inv(self.covariances[i])
                temp_cov_inv = np.linalg.inv(new_model.covariances[j])
                dist += np.sqrt(np.dot(np.dot(diff, cov_inv), diff))
                dist += np.sqrt(np.dot(np.dot(diff, temp_cov_inv), diff))
                i += 1
            distances[index] = dist
            index += 1
        return orders[np.argmin(distances)] # returns order which gives minimum net Mahalanobis distance

    def _merge(self, new_model, M):
        '''
        Merge corresponding components of new model and old model

        Refer to paper linked at the top of this file

        M is the number of samples that the new model was fit using
        '''
        order = self._match_components(new_model)
        for i in range(self.k):
            j = order[i] # get corresponding component
            old_mean = self.means[i]
            temp_mean = new_model.means[j]
            old_cov = self.covariances[i]
            temp_cov = new_model.covariances[j]
            old_weight = self.weights[i]
            temp_weight = new_model.weights[j]
            denominator = (self.N * old_weight) + (M * temp_weight) # this shows up a lot so just compute it once
            # start equation (6)
            mean = (self.N * old_weight * old_mean) + (M * temp_weight * temp_mean)
            mean /= denominator
            # start equation (7)
            cov1 = (self.N * old_weight * old_cov) + (M * temp_weight * temp_cov)
            cov1 /= denominator
            cov2 = (self.N * old_weight * old_mean * old_mean.T) + (M * temp_weight * temp_mean * temp_mean.T)
            cov2 /= denominator
            cov = cov1 + cov2 - mean * mean.T
            # check for positive-semidefinite
            cov = self._near_psd(cov)
            # start equation (8)
            weight = denominator / (self.N + M)
            # update everything
            self.means[i] = mean
            self.covariances[i] = cov
            self.weights[i] = weight

    def _near_psd(self, x):
        '''
	Calculates the nearest postive semi-definite matrix for a correlation/covariance matrix

        Code from here:
        https://stackoverflow.com/questions/10939213/how-can-i-calculate-the-nearest-positive-semi-definite-matrix
        '''
        n = x.shape[0]
        var_list = np.array([np.sqrt(x[i,i]) for i in range(n)])
        y = np.array([[x[i, j]/(var_list[i]*var_list[j]) for i in range(n)] for j in range(n)])
        while True:
            epsilon = self.epsilon
            if min(np.linalg.eigvals(y)) > epsilon:
                return x

            # Removing scaling factor of covariance matrix
            var_list = np.array([np.sqrt(x[i,i]) for i in range(n)])
            y = np.array([[x[i, j]/(var_list[i]*var_list[j]) for i in range(n)] for j in range(n)])

            # getting the nearest correlation matrix
            eigval, eigvec = np.linalg.eig(y)
            val = np.matrix(np.maximum(eigval, epsilon))
            vec = np.matrix(eigvec)
            T = 1/(np.multiply(vec, vec) * val.T)
            T = np.matrix(np.sqrt(np.diag(np.array(T).reshape((n)) )))
            B = T * vec * np.diag(np.array(np.sqrt(val)).reshape((n)))
            near_corr = B*B.T    

            # returning the scaling factors
            near_cov = np.array([[near_corr[i, j]*(var_list[i]*var_list[j]) for i in range(n)] for j in range(n)])
            if np.isreal(near_cov).all():
                break
            else:
                x = near_cov.real
        return near_cov

    def update(self, sample_array, sample_weights=None):
        '''
        Updates the model with new data without doing a full retraining.

        Parameters
        ----------
        sample_array : np.ndarray
            Array of samples to fit
        sample_weights : np.ndarray
            Weights for samples
        '''
        self.tempering_coeff /= 2
        new_model = estimator(self.k, self.max_iters, self.tempering_coeff)
        new_model.fit(sample_array, sample_weights)
        M, _ = sample_array.shape
        self._merge(new_model, M)
        self.N += M

    def score(self, sample_array, bounds=None):
        '''
        Score samples (i.e. calculate likelihood of each sample) under the current
        model.

        Parameters
        ----------
        sample_array : np.ndarray
            Array of samples to fit
        bounds : np.ndarray
            Bounds for samples, used for renormalizing scores
        '''
        n, _ = sample_array.shape
        scores = np.zeros((len(sample_array), 1))
        for i in range(self.k):
            w = self.weights[i]
            mean = self.means[i]
            cov = self.covariances[i]
            scores += np.rot90([multivariate_normal.pdf(x=sample_array, mean=mean, cov=cov, allow_singular=True)], -1) * w
            # note that allow_singular=True in the above line is probably really dumb and
            # terrible, but it seems to occasionally keep the whole thing from blowing up
            # so it stays for now
        if bounds is not None:
            # we need to renormalize the PDF
            # to do this we sample from a full distribution (i.e. without truncation) and use the
            # fraction of samples that fall inside the bounds to renormalize
            full_sample_array = self.sample(n)
            llim = np.rot90(bounds[:,[0]])
            rlim = np.rot90(bounds[:,[1]])
            n1 = np.greater(full_sample_array, llim).all(axis=1)
            n2 = np.less(full_sample_array, rlim).all(axis=1)
            normalize = np.array(np.logical_and(n1, n2)).flatten()
            m = float(np.sum(normalize)) / n
            scores /= m
        return scores

    def sample(self, n, bounds=None):
        '''
        Draw samples from the current model, either with or without bounds

        Parameters
        ----------
        n : int
            Number of samples to draw
        bounds : np.ndarray
            Bounds for samples
        '''
        sample_array = np.empty((n, self.d))
        start = 0
        for component in range(self.k):
            w = self.weights[component]
            mean = self.means[component]
            cov = self.covariances[component]
            num_samples = int(n * w)
            if component == self.k -1:
                end = n
            else:
                end = start + num_samples
            try:
                if bounds is None:
                    sample_array[start:end] = np.random.multivariate_normal(mean, cov, end - start)
                else:
                    sample_array[start:end] = truncnorm.sample(mean, cov, bounds, end - start)
                start = end
            except:
                print('Exiting due to non-positive-semidefinite')
                exit()
        return sample_array

    def print_params(self):
        '''
        Prints the model's parameters in an easily-readable format
        '''
        for i in range(self.k):
            mean = self.means[i]
            cov = self.covariances[i]
            weight = self.weights[i]
            print('________________________________________\n')
            print('Component', i)
            print('Mean')
            print(mean)
            print('Covaraince')
            print(cov)
            print('Weight')
            print(weight, '\n')



#! /usr/bin/env python
#
# util_ManualOverlapGrid.py
#
# EXAMPLES
#   util_ManualOverlapGrid.py --inj inj.xml.gz --parameter LambdaTilde  # 1d grid in changing LambdaTilde
#   util_ManualOverlapGrid.py  --verbose --parameter LambdaTilde --parameter-range '[0,1000]'
#   util_ManualOverlapGrid.py  --verbose --parameter LambdaTilde --parameter-range '[0,1000]' --parameter eta --parameter-range '[0.23,0.25]' --grid-cartesian-npts 10
#   util_ManualOverlapGrid.py --parameter s1z --parameter-range '[-0.9,0.9]' --parameter s2z --parameter-range '[-0.9,0.9]' --downselect-parameter xi --downselect-parameter-range '[-0.1,0.1]' --skip-overlap --verbose
#
# EXAMPLES WITH RANDOM FIELDS
#   python util_ManualOverlapGrid.py  --parameter mc --parameter-range [2,2.1] --random-parameter s1z --random-parameter-range [-1,1] --skip-overlap --random-parameter lambda1 --random-parameter-range [3,500] --random-parameter lambda2 --random-parameter-range [3,500]
#
# EOB SOURCE EXAMPLES
#
#   util_ManualOverlapGrid.py --inj inj.xml.gz --parameter LambdaTilde  --parameter-range '[0,1000]' --grid-cartesian-npts 10 --use-external-EOB-source
#   util_ManualOverlapGrid.py --inj inj.xml.gz --parameter LambdaTilde  --parameter-range '[0,1000]' --grid-cartesian-npts 10 --use-external-EOB-source --use-external-EOB
#   python util_ManualOverlapGrid.py --parameter s1x --parameter-range [-1,1] --parameter s1y --parameter-range [-1,1] --parameter s1z --parameter-range [-1,1] --skip-overlap  --verbose # check Kerr bound is enforced
#  
#  util_ManualOverlapGrid.py --parameter s1z --parameter-range '[-0.5,0.5]'  --parameter eta --parameter-range '[0.2,0.2499]' --verbose --seglen 16 --mass1 10 --mass2 6 --use-fisher --match-val 0.95 --approx SpinTaylorT4

# EXAMPLES: Iterative Fisher matrix creation (converging on higher match)
#   python util_ManualOverlapGrid.py  --parameter eta --parameter-range '[0.2,0.2499]' --verbose --seglen 32 --mass1 10 --mass2 6 --use-fisher --match-val 0.97 --reset-grid-via-match
#  python util_ManualOverlapGrid.py --parameter s1z --parameter-range '[-0.1,0.1]'  --parameter eta --parameter-range '[0.2,0.2499]' --verbose --seglen 16 --mass1 10 --mass2 6 --use-fisher --match-val 0.90 --approx SpinTaylorT4 --grid-cartesian-npts 500 --reset-grid-via-match
#  python util_ManualOverlapGrid.py  --parameter eta --parameter-range '[0.2,0.2499]' --verbose --seglen 32 --mass1 10 --mass2 6 --use-fisher --match-val 0.97 --reset-grid-via-match --parameter mc --parameter-range '[6.6,6.8]'
# python util_ManualOverlapGrid.py --parameter s1z --parameter-range '[-0.3,0.3]'  --parameter eta --parameter-range '[0.2,0.2499]' --verbose --seglen 16 --mass1 10 --mass2 6 --use-fisher --match-val 0.97 --approx SpinTaylorT4 --grid-cartesian-npts 500 --reset-grid-via-match --parameter mc --parameter-range '[6.6,6.8]'
# python util_ManualOverlapGrid.py --parameter s1z --parameter-range '[-0.3,0.3]'  --parameter eta --parameter-range '[0.2,0.2499]' --verbose --seglen 8 --mass1 35 --mass2 32 --use-fisher --match-val 0.97 --approx SEOBNRv2 --grid-cartesian-npts 100 --reset-grid-via-match --parameter mc --parameter-range '[28,32]'
#
# IDEA
#    - pass a list of parameters and a list of ranges
#    - signal generated on a grid using these parameters. Default layout is cartesian grid; can use others
#    - 
#
#
# ISSUES
#    - default is to regenerate the signals as needed.  Give option to archive them (very memory-painful!)
#      May want a caching interface on disk?
#    - Option to load grid from file: standard xml (injection format)
#    - default is to work for an ALIGNED-SPIN BINARY, not for more generic sources.
#
#    - enable automatic rejection of systems that violate the kerr bound

import argparse
import sys
import numpy as np
import scipy
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lalframe
import lal
import functools
import itertools

from scipy.optimize import brentq


import RIFT.physics.effectiveFisher  as eff   # for the mesh grid generation
import RIFT.physics.PrecessingFisherMatrix   as pcf   # Superior tools to perform overlaps. Will need to standardize with Evans' approach in effectiveFisher.py

from multiprocessing import Pool
try:
    import os
    n_threads = int(os.environ['OMP_NUM_THREADS'])
    print(" Pool size : ", n_threads)
except:
    n_threads=1
    print(" - No multiprocessing - ")

try:
	import NRWaveformCatalogManager3 as nrwf
	hasNR =True
except:
	hasNR=False
try:
    hasEOB=True
    import RIFT.physics.EOBTidalExternal as eobwf
except:
    hasEOB=False


###
### Linear fits. Resampling a quadratic. (Export me)
###

import RIFT.interpolators.BayesianLeastSquares

param_priors_gamma = {'s1z':0.01, 's2z': 0.01, 'xi':0.1}  # weak constraints on s1z, s2z

# def fit_quadratic(x,y,x0=None):
#     """
#     x = array so x[0] , x[1], x[2] are points.
#     """
#     x0_val = np.zeros(len(x[0]))
#     if not (x0 is None):
#         if opts.verbose:
#             print " Fisher: Using reference point ", x0
#         x0_val = x0

#     dim = len(x[0])   
#     npts = len(x)
# #    print x.shape, y.shape
#     if opts.verbose:
#         print " Fisher : dimension, npts = " ,dim, npts
#     # Constant, linear, quadratic functions. 
#     # Beware of lambda:  f_list = [(lambda x: k) for k in range(5)] does not  work, but this does
#     #     f_list = [(lambda x,k=k: k) for k in range(5)]
#     f0 = [lambda z: np.ones(len(z),dtype=np.float128)]
#     # indx_lookup_linear = {}   # protect against packing errors
#     # indx_here = len(f0)
#     # f_linear = []
#     # for k in np.arange(dim):
#     #     f_linear.append( (lambda z,k=k,x0V=x0_val: z.T[k] - x0V[k]))
#     #     indx_lookup_linear[k] =indx_here
#     #     indx_here+=1
#     f_linear = [(lambda z,k=k,x0V=x0_val: z.T[k] - x0V[k]) for k in np.arange(dim)]
#     f_quad = []
#     indx_lookup = {}
#     indx_here =len(f0)+len(f_linear) 
#     for k in np.arange(dim):
#         for q in range(k,dim):
#             f_quad.append( (lambda z,k=k,q=q: (z.T[k] - x0_val[k])*(z.T[q]-x0_val[q]))   )
#             indx_lookup[(k,q)] = indx_here
#             indx_here+=1
#     f_list=f0+f_linear + f_quad
#     n_params_model = len(f_list)
#     F = np.matrix(np.zeros((len(x), n_params_model),dtype=np.float128))
#     for q in np.arange(n_params_model):
#         fval = f_list[q](np.array(x,dtype=np.float128))
# #        print q, f_list[q], fval, fval.shape, len(x)
#         F[:,q] = np.reshape(fval, (len(x),1))
#     if opts.verbose:
#         print " ---- index pattern --- "
#         print indx_lookup
#     gamma = np.matrix( np.diag(np.ones(npts,dtype=np.float128)))
#     Gamma = F.T * gamma * F      # Fisher matrix for the fit
#     Sigma = scipy.linalg.inv(Gamma)  # Covariance matrix for the fit. WHICH CODE YOU USE HERE IS VERY IMPORTANT.
#     if opts.verbose:
#         print " -- should be identity (error here is measure of overall error) --- "
#         print "   Fisher: Matrix inversion/manipulation error ", np.linalg.norm(Sigma*Gamma - np.eye(len(Sigma))) , " which can be large if the fit coordinates are not centered near the peak"
#         print " --  --- "
#     lambdaHat =  np.array((Sigma* F.T*gamma* np.matrix(y).T))[:,0]  # point estimate for the fit parameters (i.e., fisher matrix and best fit point)
#     if opts.verbose:
#         print " Fisher: LambdaHat = ", lambdaHat
#     constant_term_est = lambdaHat[0]  # Constant term
#     linear_term_est = lambdaHat[1:dim+1]  # Coefficient of linear terms
#     my_fisher_est = np.zeros((dim,dim),dtype=np.float64)   #  A SIGNIFICANT LIMITATION...
#     for pair in indx_lookup:
#         k = pair[0]; q=pair[1];
#         indx_here = indx_lookup[pair]
#         my_fisher_est[k,q] += -lambdaHat[indx_here]
#         my_fisher_est[q,k] += -lambdaHat[indx_here]  # this will produce a factor of 2 if the two terms are identical
# #    peak_val_est = F*lambdaHat         # Peak value of the quadratic
#     if opts.verbose:
#         print "  Fisher: ", my_fisher_est
#         print "  Fisher: Sanity check (-0.5)*Fisher matrix vs components (diagonal only) : ", -0.5*my_fisher_est, "versus",  lambdaHat
#     my_fisher_est_inv = scipy.linalg.inv(my_fisher_est)   # SEE INVERSE DISCUSSION
#     if opts.verbose:
#         print " Fisher: Matrix inversion/manipulation error test 2", np.linalg.norm(np.dot(my_fisher_est,my_fisher_est_inv) - np.eye(len(my_fisher_est)))
#     # Peak value:   a - b cinv b/4
# #    print constant_term_est.shape, linear_term_est.shape, my_fisher_est_inv.shape
# #    print my_fisher_est_inv 
# #    print np.dot(my_fisher_est_inv,linear_term_est)
# #    print np.dot(linear_term_est, np.dot(my_fisher_est_inv,linear_term_est))
#     peak_val_est = float(constant_term_est) +np.dot(linear_term_est, np.dot(my_fisher_est_inv,linear_term_est))/2
#     best_val_est = x0_val +  np.dot(my_fisher_est_inv,linear_term_est)   # estimated peak location, including correction for reference point
#     if opts.verbose:
#         print " Fisher : Sanity check: peak value estimate = ", peak_val_est, " which arises as a delicate balance between ",  constant_term_est, " and ",  np.dot(linear_term_est, np.dot(my_fisher_est_inv,linear_term_est))/2
#         print " Fisher : Best coordinate estimate = ", best_val_est
#         print " Fisher : eigenvalues ", np.linalg.eig(my_fisher_est)
# #        print " Fisher : Sanity check: sizes and indexes ", dim*dim, -dim*dim+1, lambdaHat[-dim*dim+1:], len(lambdaHat[-dim*dim+1:])
#     print " WARNING: Constant offsets seen in recovery, tied to base point"
#     return [peak_val_est, best_val_est, my_fisher_est, linear_term_est]


###
### Load options
###

parser = argparse.ArgumentParser()
# Parameters
parser.add_argument("--parameter", action='append')
parser.add_argument("--parameter-range", action='append', type=str,help="Add a range (pass as a string evaluating to a python 2-element list): --parameter-range '[0.,1000.]'   MUST specify ALL parameter ranges (min and max) in order if used")
parser.add_argument("--random-parameter", action='append',help="These parameters are specified at random over the entire range, uncorrelated with the grid used for other parameters.  Use for variables which correlate weakly with others; helps with random exploration")
parser.add_argument("--random-parameter-range", action='append', type=str,help="Add a range (pass as a string evaluating to a python 2-element list): --parameter-range '[0.,1000.]'   MUST specify ALL parameter ranges (min and max) in order if used.  ")
parser.add_argument("--amplitude-order",default=-1,type=int,help="Set ampO for grid. Used in PN")
parser.add_argument("--phase-order",default=7,type=int,help="Set phaseO for grid. Used in PN")
parser.add_argument("--downselect-parameter",action='append', help='Name of parameter to be used to eliminate grid points ')
parser.add_argument("--downselect-parameter-range",action='append',type=str)
parser.add_argument("--enforce-duration-bound",default=None,type=float,help="If present, enforce a duration bound. Used to prevent grid placement for obscenely long signals, when the window size is prescribed")
parser.add_argument("--parameter-value-list", action='append', type=str,help="Add an explicit list of parameter choices to use. ONLY those values will be used. Intended for NR simulations (e.g., q, a1, a2)")
# Use external EOB for source or template?
parser.add_argument("--use-external-EOB-source",action="store_true",help="One external EOB call is performed to generate the reference signal")
parser.add_argument("--use-external-EOB",action="store_true",help="External EOB calls are performed for each template")
# Use external EOB for source or template?
parser.add_argument("--use-external-NR-source",action="store_true",help="One external NR call is performed to generate the reference signal")
parser.add_argument("--use-external-NR",action="store_true",help="External NR calls are performed for each template")
parser.add_argument("--NR-signal-group", default="Sequence-GT-Aligned-UnequalMass",help="Specific NR simulation group to use")
parser.add_argument("--NR-signal-param", default=(0.0,2.),help="Parameter value")
parser.add_argument("--NR-template-group", default=None,help="Specific NR simulation group to use")
parser.add_argument("--NR-template-param", default=None,help="Parameter value")
# Grid layout options
parser.add_argument("--use-fisher", action='store_true',help="Instead of just reporting the overlap results, perform a fit to them, and use the resulting Fisher matrix to select points. REVAMPING IMPLEMENTATION")
parser.add_argument("--fake-data", action='store_true', help="Use perfectly quadratic data. Test for fisher code.")
parser.add_argument("--uniform-spoked", action="store_true", help="Place mass pts along spokes uniform in volume (if omitted placement will be random and uniform in volume")
parser.add_argument("--linear-spoked", action="store_true", help="Place mass pts along spokes linear in radial distance (if omitted placement will be random and uniform in volume")
parser.add_argument("--grid-cartesian", action="store_true", help="Place mass points using a cartesian grid")
parser.add_argument("--grid-cartesian-npts", default=100, type=int)
parser.add_argument("--skip-overlap",action='store_true', help="If true, the grid is generated without actually performing overlaps. Very helpful for uncertain configurations or low SNR")
parser.add_argument("--reset-grid-via-match",action='store_true',help="Reset the parameter_range results so each parameter's range is limited by  match_value.  Use this ONLY for estimating the fisher matrix quickly!")
parser.add_argument("--no-reset-parameter",action='append',help="Don't reset the range of this parameter via tuning. Important for spin parameters, which can be over-tuned due to strong correlations")
parser.add_argument("--use-fisher-resampling",action='store_true',help="Resample the grid using the fisher matrix. Requires fisher matrix")
# Cutoff options
parser.add_argument("--match-value", type=float, default=0.01, help="Use this as the minimum match value. Default is 0.01 (i.e., keep almost everything)")
# Overlap options
parser.add_argument("--fisher-psd",type=str,default="SimNoisePSDaLIGOZeroDetHighPower",help="psd name (attribute in lalsimulation).  SimNoisePSDiLIGOSRD, lalsim.SimNoisePSDaLIGOZeroDetHighPower, lalsimutils.Wrapper_AdvLIGOPsd, .SimNoisePSDiLIGOSRD... ")
parser.add_argument("--psd-file",  help="File name for PSD (assumed hanford). Overrides --fisher-psd if provided")
parser.add_argument("--srate",type=int,default=16384,help="Sampling rate")
parser.add_argument("--seglen", type=float,default=256*2., help="Default window size for processing.")
parser.add_argument("--fref",type=float,default=0.);
# External grid
parser.add_argument("--use-eos", default=None, help="Equation of state to determine lambdas for given mass ranges. Filename, not EOS name (no internal database)")
parser.add_argument("--external-grid-xml", default=None,help="Inspiral XML file (injection form) for alternate grid")
parser.add_argument("--external-grid-txt", default=None, help="Cartesian grid. Must provide parameter names in header. Exactly like output of code. Last column not used.")
# Base point
parser.add_argument("--inj", dest='inj', default=None,help="inspiral XML file containing the base point.")
parser.add_argument("--event",type=int, dest="event_id", default=None,help="event ID of injection XML to use.")
parser.add_argument("--fmin", default=35,type=float,help="Mininmum frequency in Hz, default is 40Hz to make short enough waveforms. Focus will be iLIGO to keep comutations short")
parser.add_argument("--fmax",default=2000,type=float,help="Maximum frequency in Hz, used for PSD integral.")
parser.add_argument("--mass1", default=1.50,type=float,help="Mass in solar masses")  # 150 turns out to be ok for Healy et al sims
parser.add_argument("--mass2", default=1.35,type=float,help="Mass in solar masses")
parser.add_argument("--s1z", default=0.,type=float,help="Spin1z")
#parser.add_argument("--lambda1",default=590,type=float)
#parser.add_argument("--lambda2", default=590,type=float)
parser.add_argument("--eff-lambda", type=float, help="Value of effective tidal parameter. Optional, ignored if not given")
parser.add_argument("--deff-lambda", type=float, help="Value of second effective tidal parameter. Optional, ignored if not given")
parser.add_argument("--lmax", default=2, type=int)
parser.add_argument("--approx",type=str,default=None)
# Output options
parser.add_argument("--fname", default="overlap-grid", help="Base output file for ascii text (.dat) and xml (.xml.gz)")
parser.add_argument("--verbose", action="store_true",default=False, help="Extra warnings")
parser.add_argument("--extra-verbose", action="store_true",default=False, help="Lots of messages")
parser.add_argument("--save-plots",default=False,action='store_true', help="Write plots to file (only useful for OSX, where interactive is default")
opts=  parser.parse_args()

if opts.verbose:
    True
    #lalsimutils.rosDebugMessagesContainer[0]=True   # enable error logging inside lalsimutils


###
### Handle NR arguments
###
if hasNR and not ( opts.NR_signal_group in nrwf.internal_ParametersAvailable.keys()):
    if opts.NR_signal_group:
        print(" ===== UNKNOWN NR PARAMETER ====== ")
        print(opts.NR_signal_group, opts.NR_signal_param)
elif hasNR:
    if opts.NR_signal_param:
        opts.NR_signal_param = eval(str(opts.NR_signal_param)) # needs to be evaluated
    if not ( opts.NR_signal_param in nrwf.internal_ParametersAvailable[opts.NR_signal_group]):
        print(" ===== UNKNOWN NR PARAMETER ====== ")
        print(opts.NR_signal_group, opts.NR_signal_param)
if hasNR and not ( opts.NR_template_group in nrwf.internal_ParametersAvailable.keys()):
    if opts.NR_template_group:
        print(" ===== UNKNOWN NR PARAMETER ====== ")
        print(opts.NR_template_group, opts.NR_template_param)
elif hasNR:
    if opts.NR_template_param:
        opts.NR_template_param = eval(opts.NR_template_param) # needs to be evaluated
    if not ( opts.NR_template_param in nrwf.internal_ParametersAvailable[opts.NR_template_group]):
        print(" ===== UNKNOWN NR PARAMETER ====== ")
        print(opts.NR_template_group, opts.NR_template_param)




###
### Define grid overlap functions
###   - Python's 'multiprocessing' module seems to cause process lock
###

use_external_EOB=opts.use_external_EOB
Lmax = 2

def eval_overlap(grid,P_list, IP,indx):
    global opts
#    if opts.verbose: 
#        print " Evaluating for ", indx
    global use_external_EOB
    global Lmax
    global opts
    P2 = P_list[indx]
    T_here = 1./IP.deltaF
    P2.deltaF=1./T_here
#    P2.print_params()
    if not opts.skip_overlap:
        if not use_external_EOB:
            hf2 = lalsimutils.complex_hoff(P2)
        else:
            print("  Waiting for EOB waveform ....", indx, " with duration  ", T_here)
            wfP = eobwf.WaveformModeCatalog(P2,lmax=Lmax)  # only include l=2 for us.
            hf2 = wfP.complex_hoff(force_T=T_here)
        nm2 = IP.norm(hf2);  hf2.data.data *= 1./nm2
#    if opts.verbose:
#        print " Waveform normalized for ", indx
        ip_val = IP.ip(hfBase,hf2)
    line_out = []
    line_out = list(grid[indx])
    if not opts.skip_overlap:
        line_out.append(ip_val)
    else:
        line_out.append(-1)
    if opts.verbose:
        print(" Answer ", indx, line_out)
    return line_out

def calc_lambda_from_m(m, eos_fam):
    if m<10**15:
       m=m*lal.MSUN_SI

    #eos=et.read_eos(eos)
    #eos_fam=lalsim.CreateSimNeutronStarFamily(eos)

    k2=lalsim.SimNeutronStarLoveNumberK2(m, eos_fam)
    r=lalsim.SimNeutronStarRadius(m, eos_fam)

    m=m*lal.G_SI/lal.C_SI**2
    lam=2./(3*lal.G_SI)*k2*r**5
    dimensionless_lam=lal.G_SI*lam*(1/m)**5

    return dimensionless_lam


def evaluate_overlap_on_grid(hfbase,param_names, grid):
    global downselect_dict
    # Validate grid is working: Create a loop and print for each one.
    # WARNING: Assumes grid for mass-unit variables hass mass units (!)
    P_list = []
    grid_revised = []
    for line in grid:
        Pgrid = P.manual_copy()
        Pgrid.ampO=opts.amplitude_order  # include 'full physics'
        Pgrid.phaseO = opts.phase_order

        # Set attributes that are being changed as necessary, leaving all others fixed
        for indx in np.arange(len(param_names)):
            Pgrid.assign_param(param_names[indx], line[indx])

        # Downselect
        include_item =True
        if not(opts.enforce_duration_bound is None):
            if lalsimutils.estimateWaveformDuration(Pgrid)> opts.enforce_duration_bound:
                include_item = False
        for param in downselect_dict:
            if Pgrid.extract_param(param) < downselect_dict[param][0] or Pgrid.extract_param(param) > downselect_dict[param][1]:
                include_item =False
        if include_item:
         grid_revised.append(line)
         if Pgrid.m2 <= Pgrid.m1:  # do not add grid elements with m2> m1, to avoid possible code pathologies !
            P_list.append(Pgrid)
         else:
            Pgrid.swap_components()  # IMPORTANT.  This should NOT change the physical functionality FOR THE PURPOSES OF OVERLAP (but will for PE - beware phiref, etc!)
            P_list.append(Pgrid)
        else:
#            print "skipping"
#            Pgrid.print_params()
            True
#            print " skipping "
#    print "Length check", len(P_list), len(grid)
    ###
    ### Loop over grid and make overlaps : see effective fisher code for wrappers
    ###
    #  FIXME: More robust multiprocessing implementation -- very heavy!
#    p=Pool(n_threads)
    # PROBLEM: Pool code doesn't work in new configuration.
    if len(grid_revised) ==0 :
        return [],[]
    grid_out = np.array(list(map(functools.partial(eval_overlap, grid_revised, P_list,IP), np.arange(len(grid_revised)))))
    # Remove mass units at end
    for p in ['mc', 'm1', 'm2', 'mtot']:
        if p in param_names:
            indx = param_names.index(p)
            grid_out[:,indx] /= lal.MSUN_SI
    # remove distance units at end
    for p in ['distance', 'dist']:
        if p in param_names:
            indx = param_names.index(p)
            grid_out[:,indx] /= lal.PC_SI*1e6
    # Truncate grid so overlap with the base point is > opts.min_match. Make sure to CONSISTENTLY truncate all lists (e.g., the P_list)
    grid_out_new = []
    P_list_out_new = []
    for indx in np.arange(len(grid_out)):
        if opts.skip_overlap or grid_out[indx,-1] > opts.match_value:
            grid_out_new.append(grid_out[indx])
            P_list_out_new.append(P_list[indx])
    grid_out = np.array(grid_out_new)
    return grid_out, P_list_out_new



###
### Define base point 
###


# Handle PSD
# FIXME: Change to getattr call, instead of 'eval'
eff_fisher_psd = lalsim.SimNoisePSDiLIGOSRD
if not opts.psd_file:
    #eff_fisher_psd = eval(opts.fisher_psd)
    eff_fisher_psd = getattr(lalsim, opts.fisher_psd)   # --fisher-psd SimNoisePSDaLIGOZeroDetHighPower   now
    analyticPSD_Q=True
else:
    print(" Importing PSD file ", opts.psd_file)
    eff_fisher_psd = lalsimutils.load_resample_and_clean_psd(opts.psd_file, 'H1', 1./opts.seglen)
    analyticPSD_Q = False

#    from matplotlib import pyplot as plt
#    plt.plot(eff_fisher_psd.f0+np.arange(eff_fisher_psd.data.length)*eff_fisher_psd.deltaF,np.log10(eff_fisher_psd.data.data))
#    plt.show()



P=lalsimutils.ChooseWaveformParams()
if opts.inj:
    from glue.ligolw import lsctables, table, utils # check all are needed
    filename = opts.inj
    event = opts.event_id
    xmldoc = utils.load_filename(filename, verbose = True,contenthandler =lalsimutils.cthdler)
    sim_inspiral_table = table.get_table(xmldoc, lsctables.SimInspiralTable.tableName)
    P.copy_sim_inspiral(sim_inspiral_table[int(event)])
    P.fmin =opts.fmin
    if opts.approx:
        P.approx = lalsim.GetApproximantFromString(opts.approx)
        if not (P.approx in [lalsim.TaylorT1,lalsim.TaylorT2, lalsim.TaylorT3, lalsim.TaylorT4]):
            # Do not use tidal parameters in approximant which does not implement them
            print(" Do not use tidal parameters in approximant which does not implement them ")
            P.lambda1 = 0
            P.lambda2 = 0    
else:    
    P.m1 = opts.mass1 *lal.MSUN_SI
    P.m2 = opts.mass2 *lal.MSUN_SI
    P.s1z = opts.s1z
    P.dist = 150*1e6*lal.PC_SI
    if opts.eff_lambda and Psig:
        lambda1, lambda2 = 0, 0
        if opts.eff_lambda is not None:
            lambda1, lambda2 = lalsimutils.tidal_lambda_from_tilde(m1, m2, opts.eff_lambda, opts.deff_lambda or 0)
            Psig.lambda1 = lambda1
            Psig.lambda2 = lambda2

    P.fmin=opts.fmin   # Just for comparison!  Obviously only good for iLIGO
    P.ampO=opts.amplitude_order  # include 'full physics'
    P.phaseO = opts.phase_order
    if opts.approx:
        P.approx = lalsim.GetApproximantFromString(opts.approx)
        if not (P.approx in [lalsim.TaylorT1,lalsim.TaylorT2, lalsim.TaylorT3, lalsim.TaylorT4]):
            # Do not use tidal parameters in approximant which does not implement them
            print(" Do not use tidal parameters in approximant which does not implement them ")
            P.lambda1 = 0
            P.lambda2 = 0
    else:
        P.approx = lalsim.GetApproximantFromString("TaylorT4")
P.deltaT=1./16384
P.taper = lalsim.SIM_INSPIRAL_TAPER_START
P.deltaF = 1./opts.seglen #lalsimutils.findDeltaF(P)
P.fref = opts.fref
P.print_params()
Pbase = P.copy()

# Define base COMPLEX signal.  ASSUME length long enough via seglen for this  to work always
# Define base COMPLEX overlap 

hfBase = None
if opts.skip_overlap:
    print(" ---- NO WAVEFORM GENERATION ---- ")
    hfBase = None
    IP=lalsimutils.InnerProduct()  # Default, so IP.deltaF code etc does not need to be wrapped
else:
 if hasEOB and opts.use_external_EOB_source:
    print("    -------INTERFACE ------")
    print("    Using external EOB interface (Bernuzzi)   with window  ", opts.seglen)
    # Code WILL FAIL IF LAMBDA=0
    if P.lambda1<1:
        P.lambda1=1
    if P.lambda2<1:
        P.lambda2=1
    if P.deltaT > 1./16384:
        print()
    wfP = eobwf.WaveformModeCatalog(P,lmax=Lmax)  # only include l=2 for us.
    if opts.verbose:
        print(" Duration of stored signal (cut if necessary) ", wfP.estimateDurationSec())
    hfBase = wfP.complex_hoff(force_T=opts.seglen)
    print("EOB waveform length ", hfBase.data.length)
    print("EOB waveform duration", -hfBase.epoch)
 elif opts.use_external_EOB_source and not hasEOB:
    # do not do something else silently!
    print(" Failure: EOB requested but impossible ")
    sys.exit(0)
 elif opts.use_external_NR_source and hasNR:
    m1Msun = P.m1/lal.MSUN_SI;     m2Msun = P.m2/lal.MSUN_SI
    if m1Msun < 50 or m2Msun < 50:
        print(" Invalid NR mass ")
        sys.exit(0)
    print(" Using NR ", opts.NR_signal_group, opts.NR_signal_param)
    T_window = 16. # default 
    wfP = nrwf.WaveformModeCatalog(opts.NR_signal_group, opts.NR_signal_param, clean_initial_transient=True,clean_final_decay=True, shift_by_extraction_radius=True, lmax=opts.lmax,align_at_peak_l2_m2_emission=True,build_strain_and_conserve_memory=True)
    q = wfP.P.m2/wfP.P.m1
    print(" NR q  (overrides anything)", q)
    mtotOrig  =(wfP.P.m1+wfP.P.m2)/lal.MSUN_SI
    wfP.P.m1 *= (m1Msun+m2Msun)/mtotOrig
    wfP.P.m2 *= (m1Msun+m2Msun)/mtotOrig

    wfP.P.deltaT = 1./opts.srate
    print(" NR duration (in s) of simulation at this mass = ", wfP.estimateDurationSec())
    print(" NR starting 22 mode frequency at this mass = ", wfP.estimateFminHz())
    T_window = max([16, 2**int(np.log(wfP.estimateDurationSec())/np.log(2)+1)])
    wfP.P.deltaF = 1./T_window
    print(" Final T_window ", T_window)
    wfP.P.radec = False  # use a real source with a real instrument
    wfP.P.fmin = 10
    print("  ---- NR interface: Overriding parameters to match simulation requested ---- ")
    wfP.P.print_params()
    hfBase = wfP.complex_hoff(force_T=T_window)
 elif opts.use_external_NR_source and not hasNR:
    print(" Failure: NR requested but impossible ")
    sys.exit(0)
 else:
    print("    -------INTERFACE ------")
    print("    Using lalsuite   ", hasEOB, opts.use_external_EOB_source)
    hfBase = lalsimutils.complex_hoff(P)
 IP = lalsimutils.CreateCompatibleComplexOverlap(hfBase,analyticPSD_Q=analyticPSD_Q,psd=eff_fisher_psd,fMax=opts.fmax,interpolate_max=True)
 nmBase = IP.norm(hfBase)
 hfBase.data.data *= 1./nmBase
 if opts.verbose:
    print(" ------  SIGNAL DURATION ----- ")
    print(hfBase.data.length*P.deltaT)

###
### Define parameter ranges to be changed
###

if not opts.parameter:
    param_names =[ 'eta', 'LambdaTilde']  # override options for now
    param_ranges =[ [0.23, 0.25], [0, 1000]]
    pts_per_dim = [ 10,10]
elif not(opts.parameter is None):
    param_names = opts.parameter
    for param in param_names:
        # Check if in the valid list
        if not(param in lalsimutils.valid_params):
            print(' Invalid param ', param, ' not in ', lalsimutils.valid_params)
            sys.exit(0)
    npts_per_dim = int(np.power(opts.grid_cartesian_npts, 1./len(param_names)))+1
    if opts.downselect_parameter:
        npts_per_dim *= np.power(2*len(opts.downselect_parameter), 1/len(param_names))  # if we provide cutoffs, increase the underlying grid size. Make SURE we get target point size!
    pts_per_dim = npts_per_dim*np.ones(len(param_names))  # ow!
    param_ranges = []
    if len(param_names) == len(opts.parameter_range):
        param_ranges = list(map(eval, opts.parameter_range))
        # Rescale hand-specified ranges to SI units
        for p in ['mc', 'm1', 'm2', 'mtot']:
          if p in param_names:
            indx = param_names.index(p)
            #print p, param_names[indx], param_ranges[indx]
            param_ranges[indx]= np.array(param_ranges[indx])* lal.MSUN_SI
        # Rescale hand-specified ranges to SI units
        for p in ['distance','dist']:
          if p in param_names:
            indx = param_names.index(p)
            #print p, param_names[indx], param_ranges[indx]
            param_ranges[indx]= np.array(param_ranges[indx])* lal.PC_SI*1e6
    else:
     for param in param_names:
        if param == 'mc':
            val_center = P.extract_param(param)
            param_ranges.append( [val_center*0.99, val_center*1.01])
        elif param == 'eta':
            val_center = P.extract_param(param)
            mcval = P.extract_param('mc')
            # Use range so limited to smaller body > 1 Msun: HARDCODED
            eta_min = lalsimutils.eta_crit(mcval, 1)
            param_ranges.append( [eta_min, 0.25])
        elif param == 'delta_mc':
            param_ranges.append( [0, 1])
        # FIXME: Implement more parameter ranges, using a lookup-based method
        elif param == 'LambdaTilde':
            #val_center = P.extract_param(param)
            param_ranges.append( [0, 1000]) # HARDCODED
        else:
            print(" Parameter not implemented ", param)
            sys.exit(0)

    if opts.verbose:
        print(" ----- PARAMETER RANGES ----- ")
        for indx in np.arange(len(param_names)):
            print(param_names[indx], param_ranges[indx], pts_per_dim[indx])


template_min_freq = opts.fmin
ip_min_freq = opts.fmin


###
### Auto-tune parameter range based on match threshold
###
if not(opts.skip_overlap) and opts.reset_grid_via_match and opts.match_value <1:
    # Based on the original effective fisher code: 'find_effective_Fisher_region'
    TOL=(1-opts.match_value)*1e-2
    maxit=50
#    TOL=1e-5
    for indx in np.arange(len(param_names)):
        PT = Pbase.copy()  # same as the grid, but we will reset all its parameters
        param_now = param_names[indx]
        if not(opts.no_reset_parameter is None):
          if param_now in opts.no_reset_parameter:
            print(" ==> not retuning range for ", param_now)
            continue
        param_peak = Pbase.extract_param(param_now)
        fac_print =1.0
        if param_now in ['m1', 'm2', 'mc']:
            fac_print = lal.MSUN_SI
        if opts.verbose:
            print(" Optimizing for ", param_now, " with peak expected at value = ", param_peak)
            PT.print_params()
        def ip_here(x):
            PT.assign_param(param_now,x)
            PT.deltaF = IP.deltaF
            hf_now = lalsimutils.complex_hoff(PT)
            nm_now = IP.norm(hf_now)
            val = IP.ip(hfBase,hf_now)/nm_now
            if opts.verbose:
                print(param_now, x, val)
            return val - opts.match_value
        def ip_here_squared(x):
            val= ip_here(x)
            return val**2
        try:
            print(" Minimum: looking between ",param_ranges[indx][0]/fac_print,param_peak/fac_print, " delta ", np.abs(param_peak - param_ranges[indx][0])/fac_print)
#            if np.abs(param_peak - param_ranges[indx][0])/fac_print < 1e-2:  # very narrow placement range. Want to avoid jumping to the wrong side1
#                print " Using minimization code ...  "
#                
#            else:
            param_min = brentq(ip_here,param_ranges[indx][0],param_peak ,xtol=TOL,maxiter=maxit)
            if param_min > param_peak:
                print(" Ordering problem, using minimization code as backup ")
                param_min = brent(ip_here_squared, brack=(param_ranges[indx][0],param_peak, param_ranges[indx][1]), tol=TOL)
        except:
            print("  Range retuning: minimum for ", param_now)
            param_min = param_ranges[indx][0]
        try:
            print(" Maximum: looking between ",param_peak/fac_print, param_ranges[indx][1]/fac_print)
            param_max = brentq(ip_here,param_peak, param_ranges[indx][1],xtol=TOL)
        except:
            print("  Range retuning: maximum for ", param_now)
            param_max = param_ranges[indx][1]
        if np.abs(param_max - param_min)/(np.abs(param_max)+np.abs(param_min)) < 1e-6:  # override if we have catastrophically close placement
            print(" Override: tuned parameters got too close, returning to original range ")
            param_min = param_ranges[indx][0]
            param_max = param_ranges[indx][1]
        print(" Revised range for parameter ", np.array(param_ranges[indx])/fac_print, " to ", [param_min/fac_print,param_max/fac_print], " around ", param_peak/fac_print)
        param_ranges[indx][0] = param_min
        param_ranges[indx][1] = param_max


###
### Downselect parameters
###



downselect_dict = {}

# Enforce Kerr bound
downselect_dict['chi1'] = [0,1]
downselect_dict['chi2'] = [0,1]
for param in ['s1z', 's2z', 's1x','s2x', 's1y', 's2y']:
    downselect_dict[param] = [-1,1]
# Enforce definition of eta
downselect_dict['eta'] = [0,0.25]

if opts.downselect_parameter:
    dlist = opts.downselect_parameter
    dlist_ranges  = list(map(eval,opts.downselect_parameter_range))
else:
    dlist = []
    dlist_ranges = []
if len(dlist) != len(dlist_ranges):
    print(" downselect parameters inconsistent", dlist, dlist_ranges)
for indx in np.arange(len(dlist_ranges)):
    downselect_dict[dlist[indx]] = dlist_ranges[indx]


print(" Downselect dictionary (before scaling) ", downselect_dict)

# downselection procedure: fix units so I can downselect on mass parameters
for p in ['mc', 'm1', 'm2', 'mtot']:
    if p in downselect_dict.keys():
        downselect_dict[p] = np.array(downselect_dict[p],dtype=np.float64)
        downselect_dict[p] *= lal.MSUN_SI  # add back units

print(" Downselect dictionary (with units) ", downselect_dict)


###
### Prior dictionary on coordinates (mainly for Fisher resampling).
###
prior_dict = {}
prior_dict['s1z'] =  2  # provide std dev. Don't want to allow arbitrarily large spins
prior_dict['s2z'] =  2  # provide std dev. Don't want to allow arbitrarily large spins
prior_dict['xi']  = 2    
prior_dict['chieff_aligned']  = 2    
prior_dict['eta'] = 1    # provide st dev. Don't want to allow arbitrary eta.


###
### Lay out grid, currently CARTESIAN.   OPTIONAL: Load grid from file
###
# WARNINGS: The code will NOT enforce sanity, and can produce lambda's < 0. This may cause some codes to barf.
# FIXME: Use seed cartesian grid to help lay out an effective Fisher grid


# Base Cartesian grid
print(param_ranges,pts_per_dim)
grid_tuples = eff.make_regular_1d_grids(param_ranges, pts_per_dim)
#print "  NEED TO IMPLEMENT: Stripping of unphysical parameters "
grid = eff.multi_dim_grid(*grid_tuples)  # each line in 'grid' is a set of parameter values
print(grid.shape)

# Extend to use alternative parameters
if not (opts.random_parameter is None) and not(opts.parameter is None):
    indx_base = len(opts.parameter)
#    print param_names, param_ranges
    param_names += opts.random_parameter
    param_ranges += list(map(eval, opts.random_parameter_range))  # do not randomize mass or distance parameters, not being rescaled
    print(param_names, param_ranges,indx_base)
    grid_extra = np.zeros( (len(grid),len(opts.random_parameter)) )
#    print grid_extra.shape
    for indx in np.arange(len(opts.random_parameter)):
        range_here = np.array(param_ranges[ indx_base+indx])
        if param_names[indx_base+indx] in ['mc','mtot','m1','m2']:
            range_here *= lal.MSUN_SI
        grid_extra[:,indx] = np.random.uniform( range_here[0],range_here[1],size=len(grid))

    grid = np.hstack((grid,grid_extra))

elif opts.parameter is None:
    param_names =  opts.random_parameter
    param_ranges = list(map(eval, opts.random_parameter_range))
    print(param_names, param_ranges)
    grid = np.zeros( (opts.grid_cartesian_npts,len(opts.random_parameter)) )
    for indx in np.arange(len(opts.random_parameter)):
        range_here = np.array(param_ranges[ indx])
        if param_names[indx] in ['mc','mtot','m1','m2']:
            range_here *= lal.MSUN_SI
        grid[:,indx] = np.random.uniform( range_here[0],range_here[1],size=len(grid))


# Special check: m2<m1 : if both names appear, strip parameters from the grid
if ('m1' in param_names) and ('m2' in param_names):
    print(" Grid uses m1, m2. Eliminating points with m1<m2")
    indx1 = param_names.index('m1')
    indx2 = param_names.index('m2')
    indxOk = grid[:,indx1] >= grid[:,indx2]
    grid = grid[indxOk] # Boolean indexing
    print(" Revised grid size : ", len(grid))

# If external grid provided, erase this grid and set of names, and replace it with the new one.
if opts.external_grid_txt:
    tmp = np.genfromtxt(opts.external_grid_txt, names=True)
    raw_names = tmp.dtype.names
#    print tmp, tmp['eta'], tmp['ip'], raw_names
    param_names = np.array(list(set(raw_names) - set(['ip'])))
    for p in ['mc', 'm1', 'm2', 'mtot']:
        if p in param_names:
            tmp[p] *= lal.MSUN_SI  # add back units
    print(param_names)
    grid = np.array(tmp[param_names])
    param_names = list(param_names)
    # rescale arrays that correspond to masses

#    print grid, grid[1][0]
#    sys.exit(0)


#if using an external EOS add lambda to grid (Richard, you may want to fix this to be more general)
elif opts.use_eos!=None:
#   from gwemlightcurves.KNModels import table 
   import EOSManager

   grid_tmp=np.zeros((len(grid[:,0]), len(grid[0,:])+2))
   anEOS = EOSManager.EOSLALSimulation(opts.use_eos)
#   eos,eos_fam=table.get_lalsim_eos(opts.use_eos)
   eos_fam = anEOS.eos_fam

   for i in range(0,len(grid[0,:])):
       grid_tmp[:,i]=grid[:,i]

   param_names.append('lambda1')
   param_names.append('lambda2')

   mc_indx=param_names.index('mc')
   my_transform = lambda x: x
   if 'eta' in param_names:
       eta_indx=param_names.index('eta')
   else:
       eta_indx = param_names.index('delta_mc')
       my_transform = lambda x: 0.25*(1.-x*x)
   lam1_indx=param_names.index('lambda1')
   lam2_indx=param_names.index('lambda2')

   for i in range(0,len(grid[:,mc_indx])): # Ridiculously inefficient
       # fail to assign anything if m1 or m2 is out of range
       m1=lalsimutils.mass1(grid[i,mc_indx],my_transform(grid[i,eta_indx]))
       if m1/lal.MSUN_SI < anEOS.mMaxMsun:
           grid_tmp[i,lam1_indx]= anEOS.lambda_from_m(m1)  # calc_lambda_from_m(m1,eos_fam)
       m2=lalsimutils.mass2(grid[i,mc_indx],my_transform(grid[i,eta_indx]))
       if m2/lal.MSUN_SI < anEOS.mMaxMsun:
           grid_tmp[i,lam2_indx]= anEOS.lambda_from_m(m2)  #calc_lambda_from_m(m2,eos_fam)

   # Remove points with zero lambda1 or lambda2?  
   #     Not necessarily ... think about it ... but require both for now
   indx_ok  = np.logical_and(grid_tmp[:,lam1_indx]>0 ,grid_tmp[:,lam2_indx]>0 )
   grid=grid_tmp[indx_ok]


grid_out, P_list = evaluate_overlap_on_grid(hfBase, param_names, grid)
if len(grid_out)==0:
    print(" No points survive....")

###
### (Fisher matrix-based grids): 
###     - Use seed cartesian grid to compute the effective fisher matrix
###     - Loop *again* to evaluate overlap on that grid
###


if opts.use_fisher:
    rho_fac = 8
    if opts.fake_data:
        grid_out[:,-1] = np.ones(len(grid_out))
        for k in np.arange(len(param_names)):
            grid_out[:,-1] += -1.0*np.power(grid_out[:,k] - np.mean(grid_out[:,k]),2)/np.power(np.std(grid_out[:,k]),2)   # simple quadratic sum in all data. Change to rhange

    # Save data for fisher
    headline = ' '.join(param_names + ['ip'])
    if int(np.version.short_version.split('.')[1])>=7:  # http://docs.scipy.org/doc/numpy/reference/generated/numpy.savetxt.html
        np.savetxt(opts.fname+"_fisher_input.dat", grid_out, header=headline)
    else:
        np.savetxt(opts.fname+"_fisher_input.dat", grid_out)   # 


    # Reference point for fit should NOT MATTER
    x0_val_here =grid_out[0,:len(param_names)]
#    print grid_out[0], x0_val_here
    print(" Generating Fisher matrix using N = ", len(grid_out), " surviving points with match > ", opts.match_value)
    print(" Creating nominal prior on parameters ")
    prior_x_gamma = np.zeros( (len(param_names),len(param_names)) )
    for indx in np.arange(len(param_names)):
        if param_names[indx] in prior_dict:
            prior_x_gamma[indx,indx] = 1./prior_dict[param_names[indx]]**2 /(rho_fac*rho_fac) # must divide, because we work with the scaled Fisher
    the_quadratic_results = BayesianLeastSquares.fit_quadratic( grid_out[:,:len(param_names)], grid_out[:,len(param_names)],x0=x0_val_here,prior_x_gamma=prior_x_gamma)#x0=None)#x0_val_here)
    print("Fisher matrix results (raw) :", the_quadratic_results)
    peak_val_est, best_val_est, my_fisher_est, linear_term_est,fn_estimate = the_quadratic_results
    np.savetxt("fisher_reference.dat",x0_val_here,header=' '.join(param_names)) 
    np.savetxt("fisher_peakval.dat",[peak_val_est])   # generally not very useful
    np.savetxt("fisher_bestpt.dat",best_val_est,header=' '.join(param_names))  
    np.savetxt("fisher_gamma.dat",my_fisher_est,header=' '.join(param_names))
    np.savetxt("fisher_linear.dat",linear_term_est)

    my_eig= scipy.linalg.eig(my_fisher_est)
    if any(np.real(my_eig[0]) < 0) : 
        print(" Negative eigenvalues COULD preclude resampling ! Use a prior to regularize")
        print(" Eigenvalue report ", my_eig)
        print(" HOPE that priors help !")
        np.savetxt("fisher_NEG",my_eig[0].flatten())  # filename with eigenvalues

        # As a stopgap, save results with *regularized* fisher
        print(" ==> Regularizing fisher matrix results <== ")
        the_quadratic_results = BayesianLeastSquares.fit_quadratic( grid_out[:,:len(param_names)], grid_out[:,len(param_names)],x0=x0_val_here,prior_x_gamma=prior_x_gamma,hard_regularize_negative=True)#x0=None)#x0_val_here)
        peak_val_est, best_val_est, my_fisher_est, linear_term_est,fn_estimate = the_quadratic_results
        np.savetxt("fisher_gamma.dat",my_fisher_est,header=' '.join(param_names))

#        sys.exit(0)

    if opts.use_fisher and opts.use_fisher_resampling:  # dump real data, NOT faked data. You should ALWAYS enter this logic tree
     # this grid will have mass units ! Possibly catastrophic?
     npts_out = opts.grid_cartesian_npts*10  # Use more points in the search grid than the small grid used for Fisher. Aim for 1k points
     grid_fisher = -1*np.ones((npts_out,len(param_names)+1))
     prior_gamma= np.zeros(my_fisher_est.shape)
     for indx in np.arange(len(param_names)):
         if param_names[indx] in param_priors_gamma:
             prior_gamma[indx][indx]  = param_priors_gamma[param_names[indx]]
     print(" Using prior matrix to generate candidate draws from likelihood: ", prior_gamma)

     grid_fisher[:,:len(param_names)] = BayesianLeastSquares.fit_quadratic_and_resample(grid_out[:,:len(param_names)], grid_out[:,len(param_names)],rho_fac=8,npts=npts_out,x0=x0_val_here,prior_quadratic_gamma=prior_gamma,hard_regularize_negative=True)
     grid_fisher[:,-1] = -1*np.ones(npts_out)
     # Convert grid to physical systems.  Drop systems which make no sense
     P_list = []
     grid_revised = []
     for line in grid_fisher:
        Pgrid = P.manual_copy()
        include_item =True
        # Set attributes that are being changed as necessary, leaving all others fixed
        for indx in np.arange(len(param_names)):
            if param_names[indx] in downselect_dict:
                if line[indx] < downselect_dict[param_names[indx]][0] or line[indx] > downselect_dict[param_names[indx]][1]:
                    include_item = False
                    if opts.extra_verbose:
                        print(" Skipping " , line)
            # if parameter involes a mass parameter, scale it to sensible units
            fac = 1
            if param_names[indx] in ['mc', 'mtot', 'm1', 'm2']:
                fac = lal.MSUN_SI
            # do assignment of parameters anyways, as we will skip it momentarily
            Pgrid.assign_param(param_names[indx], line[indx]*fac)

        # Downselect.
        for param in downselect_dict:
             if Pgrid.extract_param(param) < downselect_dict[param][0] or Pgrid.extract_param(param) > downselect_dict[param][1]:
                 if opts.extra_verbose:
                     print(" Skipping " , line)
                 include_item =False
        if include_item:
         grid_revised.append(line)
         if Pgrid.m2 <= Pgrid.m1:  # do not add grid elements with m2> m1, to avoid possible code pathologies !
            P_list.append(Pgrid)
         else:
            Pgrid.swap_components()  # IMPORTANT.  This should NOT change the physical functionality FOR THE PURPOSES OF OVERLAP (but will for PE - beware phiref, etc!)
            P_list.append(Pgrid)
        else:
            True

     print(" Original grid size", len(grid_out))
     grid_out=[]  # force reset of this variable..no crap inherited
     grid_out = np.array(grid_revised)
     print(" Revised grid size", len(grid_out))

if opts.linear_spoked or opts.uniform_spoked:
    print(" Effective fisher report. GRID NOT YET IMPLEMENTED ")
    if len(param_names)==2:
        fitgamma = eff.effectiveFisher(eff.residuals2d, grid_out[:,-1], *grid_out[:,0:len(param_names)-1])
        gam = eff.array_to_symmetric_matrix(fitgamma)
        evals, evecs, rot = eff.eigensystem(gam)
        # Print information about the effective Fisher matrix
        # and its eigensystem
        print("Least squares fit finds ", fitgamma)
        print("\nFisher matrix:")
        print("eigenvalues:", evals)
        print("eigenvectors:")
        print(evecs)
        print("rotation ")
        print(rot)

    else:
        print(" Higher-dimensional grids not yet implemented ")
        sys.exit(0)
    print("Fisher grid not yet implemented")
    sys.exit(0)
                             


###
### Optional: Write grid to XML file (ONLY if using cutoff option)
###
n_out = np.min([opts.grid_cartesian_npts,len(P_list)])
lalsimutils.ChooseWaveformParams_array_to_xml(P_list[:n_out], fname=opts.fname, fref=P.fref)
print("  ==> Writing npts = ", n_out, " out of ", len(P_list))

###
### Write output to text file:  p1 p2 p3 ... overlap, only including named params
###
headline = ' '.join(param_names + ['ip'])
if int(np.version.short_version.split('.')[1])>=7:  # http://docs.scipy.org/doc/numpy/reference/generated/numpy.savetxt.html
    np.savetxt(opts.fname+".dat", grid_out, header=headline)
else:
    np.savetxt(opts.fname+".dat", grid_out)   # 


###
### Optional: Scatterplot
###
if opts.save_plots and opts.verbose and len(param_names)==1 and len(grid_out)>0:
    import matplotlib.pyplot as plt
    fig = plt.figure()
    plt.plot(grid_out[:,0], grid_out[:,1])
    plt.savefig("fig-grid2d.png")
#    plt.show()

if opts.save_plots and opts.verbose and len(param_names)==2:
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(grid_out[:,0], grid_out[:,1], grid_out[:,2])
    plt.savefig("fig-grid3d.png")
#    plt.show()

print(" ---- DONE ----")



#! /bin/bash
# Very very quick tool to yank out posterior quantiles from an ascii file
#  Example

#  util_CharacterizePosterior.py --parameter mc --quantiles '[0.9]'
# for i in S*.dat; do echo $i; python util_CharacterizePosterior.py --fname $i --parameter mc --quantiles [0.05,0.95] --parameter q; done



import argparse
import numpy as np
import RIFT.lalsimutils


parser = argparse.ArgumentParser()
parser.add_argument("--fname",type=str,help="filename of *.dat file [standard ILE output]")
parser.add_argument("--parameter",action='append',help="name of param")
parser.add_argument("--quantiles",type=str,help="quantiles")
opts=  parser.parse_args()

#print opts.parameter

quantile_list = np.array(eval(opts.quantiles))
#print opts.quantiles, quantile_list

dat = np.genfromtxt(opts.fname,names=True)
for param in opts.parameter:
    # lightweight conversion: not full capability
    if param in dat.dtype.names:
        dat_1d = dat[param]
    elif param == 'mc':
        dat_1d = RIFT.lalsimutils.mchirp(dat['m1'],dat['m2'])
    elif param == 'q':
        dat_1d = dat['m2']/dat['m1']
    elif param == 'xi':
        dat_1d = (dat['m2']*dat['a2z']+dat['m1']*dat['a1z'])/(dat['m1']+dat['m2'])
    quant_here  = np.percentile(dat_1d,100*quantile_list)
    print(param, ' '.join(map(str,quant_here)))



#! /usr/bin/env python
#  
#
# Tool: dumps a candidate waveform to a frame file.
# Default GPS time used is boring
# Tries to be as close to C as possible -- no interface via pylal/glue
#
# EXAMPLE
#   python util_NRWriteFrame.py --group 'Sequence-SXS-All' --param 1 --verbose
#   python util_NRWriteFrame.py --incl 1.5 --verbose     # edge on
#
#  WARNING: My version does NOT interpolate the signal to a synchronized set of sample times.
#                       This may cause problems for some applications, particularly at low sample rates.

from __future__ import print_function

import argparse
import sys
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lalframe
import lal

import RIFT.physics.EOBTidalExternalC as eobwf


parser = argparse.ArgumentParser()
parser.add_argument("--fname", default=None, help = "Base name for output frame file. Otherwise auto-generated ")
parser.add_argument("--instrument", default="H1",help="Use H1, L1,V1")
parser.add_argument("--inj", dest='inj', default=None,help="inspiral XML file containing injection information. Used for extrinsic information only")
parser.add_argument("--mass1", default=1.50,type=float,help="Mass in solar masses")  # 150 turns out to be ok for Healy et al sims
parser.add_argument("--mass2", default=1.35,type=float,help="Mass in solar masses")
parser.add_argument("--lambda1",default=590,type=float)
parser.add_argument("--lambda2", default=590,type=float)
parser.add_argument("--fmin", default=35,type=float,help="Mininmum frequency in Hz, default is 40Hz to make short enough waveforms. Focus will be iLIGO to keep comutations short")
parser.add_argument("--lmax", default=2, type=int)
parser.add_argument("--event",type=int, dest="event_id", default=None,help="event ID of injection XML to use.")
parser.add_argument("--srate",type=int,default=16384,help="Sampling rate")
parser.add_argument("--seglen", type=float,default=256*2., help="Default window size for processing.")
parser.add_argument("--incl",default=0,type=float,help="Set the inclination of the simuation. Helpful for aligned spin tests")
parser.add_argument("--start", type=int,default=None)
parser.add_argument("--stop", type=int,default=None)
parser.add_argument("--approx",type=str,default=None,help="Unused")
parser.add_argument("--single-ifo",action='store_true',default=None,help="Unused")
parser.add_argument("--verbose", action="store_true",default=False, help="Required to build post-frame-generating sanity-test plots")
parser.add_argument("--save-plots",default=False,action='store_true', help="Write plots to file (only useful for OSX, where interactive is default")
opts=  parser.parse_args()



# Window size : 8 s is usually more than enough, though we will fill a 16s buffer to be sure.
T_window = 128. # default.  Note in practice most NS-NS waveforms will need to be many tens of minutes long 



# Generate signal
P=lalsimutils.ChooseWaveformParams()
P.m1 = opts.mass1 *lal.MSUN_SI
P.m2 = opts.mass2 *lal.MSUN_SI
P.dist = 150*1e6*lal.PC_SI
P.lambda1  = opts.lambda1
P.lambda2  = opts.lambda2
P.fmin=opts.fmin   # Just for comparison!  Obviously only good for iLIGO
P.ampO=-1  # include 'full physics'
P.deltaT=1./16384
P.taper = lalsim.SIM_INSPIRAL_TAPER_START
# This must be done BEFORE changing the duration
P.scale_to_snr(20,lalsim.SimNoisePSDaLIGOZeroDetHighPower,['H1', 'L1'])
if opts.start and opts.stop:
    opts.seglen = opts.stop-opts.start # override
P.deltaF = 1./opts.seglen #lalsimutils.findDeltaF(P)
if P.deltaF > 1./T_window:
    print(" time too short ")


if not opts.inj:
    P.taper = lalsimutils.lsu_TAPER_START
    P.tref = 1000000000 #1000000000            # default
    P.dist = 150*1e6*lal.PC_SI # default
else:
    from glue.ligolw import lsctables, table, utils # check all are needed

    filename = opts.inj
    event = opts.event_id
    xmldoc = utils.load_filename(filename, verbose = True,contenthandler =lalsimutils.cthdler)
    sim_inspiral_table = table.get_table(xmldoc, lsctables.SimInspiralTable.tableName)
    P.copy_sim_inspiral(sim_inspiral_table[int(event)])
P.detector = opts.instrument

P.print_params()

# FAIL if masses are not viable
if P.m1/lal.MSUN_SI > 3 or P.m2/lal.MSUN_SI > 3:
    print(" Invalid NS mass ")
    sys.exit(0)

wfP = eobwf.WaveformModeCatalog(P,lmax=opts.lmax)
print(" Loaded modes ", wfP.waveform_modes_complex.keys())
print(" Duration of stored signal ", wfP.estimateDurationSec())
mtotMsun = (wfP.P.m1+wfP.P.m2)/lal.MSUN_SI


# Generate signal
hoft = wfP.real_hoft()   # include translation of source, but NOT interpolation onto regular time grid
print(" Original signal (min, max) ", np.min(hoft.data.data) ,np.max(hoft.data.data))
print(" Original signal duration ", hoft.deltaT*hoft.data.length)
# zero pad to be opts.seglen long
TDlenGoal = int(opts.seglen/hoft.deltaT)
if TDlenGoal < hoft.data.length:
    print(" seglen too short -- signal truncation would be required")
    sys.exit(0)
nptsOrig = hoft.data.length
hoft = lal.ResizeREAL8TimeSeries(hoft, 0, TDlenGoal)
hoft.data.data[nptsOrig:TDlenGoal] = 0 # np.zeros(TDlenGoal-nptsOrig) # zero out the tail
print(" Resized signal (min, max) ", np.min(hoft.data.data) ,np.max(hoft.data.data))
# zero pad some more on either side, to make sure the segment covers start to stop
if opts.start and hoft.epoch > opts.start:
    nToAddBefore = int((hoft.epoch-opts.start)/hoft.deltaT)
    print("Padding start ", nToAddBefore, hoft.data.length)
    ht = lal.CreateREAL8TimeSeries("Template h(t)", 
            hoft.epoch - nToAddBefore*hoft.deltaT, 0, hoft.deltaT, lalsimutils.lsu_DimensionlessUnit, 
            hoft.data.length+nToAddBefore)
    ht.data.data[0:ht.data.length] = np.zeros(ht.data.length)  # initialize to zero for safety
    ht.data.data[nToAddBefore:nToAddBefore+hoft.data.length] = hoft.data.data
    hoft = ht

if opts.stop and hoft.epoch+hoft.data.length*hoft.deltaT < opts.stop:
    nToAddAtEnd = int( (-(hoft.epoch+hoft.data.length*hoft.deltaT)+opts.stop)/hoft.deltaT)
else:
    nToAddAtEnd=0
if nToAddAtEnd <=0:
    nToAddAtEnd = int(1/hoft.deltaT)  # always at at least 1s of padding at end
print("Padding end ", nToAddAtEnd, hoft.data.length)
nptsNow = hoft.data.length
hoft = lal.ResizeREAL8TimeSeries(hoft,0, int(hoft.data.length+nToAddAtEnd))
hoft.data.data[nptsNow:hoft.data.length] = 0
print(" Padded signal (min, max) ", np.min(hoft.data.data) ,np.max(hoft.data.data))


channel = opts.instrument+":FAKE-STRAIN"

tstart = int(hoft.epoch)
duration = int(hoft.data.length*hoft.deltaT)
if not opts.fname:
    fname = opts.instrument.replace("1","")+"-fake_strain-"+str(tstart)+"-"+str(duration)+".gwf"

print("Writing signal with ", hoft.data.length*hoft.deltaT, " to file ", fname)
print("Maximum original ", np.max(hoft.data.data))
print("Start time", hoft.epoch)
lalsimutils.hoft_to_frame_data(fname,channel,hoft)

bNoInteractivePlots=True # default
fig_extension = '.jpg'
try:
    import matplotlib
    print(" Matplotlib backend ", matplotlib.get_backend())
    if matplotlib.get_backend() is 'MacOSX':
        if opts.save_plots:
            print("  OSX without interactive plots")
            bNoInteractivePlots=True
            fig_extension='.jpg'
        else:  #  Interactive plots
            print("  OSX with interactive plots")
            bNoInteractivePlots=False
    elif matplotlib.get_backend() is 'agg':
        fig_extension = '.png'
        bNoInteractivePlots=True
        print(" No OSX; no interactive plots ")
    else:
        print(" Unknown configuration ")
        fig_extension = '.png'
        bNoInteractivePlots =True
    from matplotlib import pyplot as plt
    bNoPlots=False
except:
    from matplotlib import pyplot as plt
    fig_extension = '.png'
    print(" - no matplotlib - ")
    bNoInteractivePlots = True
    bNoPlots = False


# TEST: Confirm it works by reading the frame
if opts.verbose and not bNoPlots:
    import os
#    from matplotlib import pyplot as plt
    # First must create corresponding cache file
    os.system("echo "+ fname+ " | lalapps_path2cache   > test.cache")
    # Now I can read it
    # Beware that the results are OFFSET FROM ONE ANOTHER due to PADDING,
    #    but that the time associations are correct
    hoft2 = lalsimutils.frame_data_to_hoft("test.cache", channel)
    tvals2 = (float(hoft2.epoch) - float(wfP.P.tref)) +  np.arange(hoft2.data.length)*hoft2.deltaT
    tvals = (float(hoft.epoch) - float(wfP.P.tref)) +  np.arange(hoft.data.length)*hoft.deltaT

    ncrit = np.argmax(hoft2.data.data)
    tcrit = float(hoft2.epoch) - float(wfP.P.tref) + ncrit*hoft2.deltaT    # zero time

    print(" Maximum original ", np.max(hoft.data.data), " size ", len(tvals), len(hoft.data.data))
    print(" Maximum frames ", np.max(hoft2.data.data), " size ", len(tvals2), len(hoft2.data.data))
    print(" Location of maximum in samples. relative time ", ncrit, tcrit)
    print(" Location of maximum in samples, compared to tref", tcrit+P.tref, end=' ')
    print(" Location of maximum as GPS time ", ncrit*hoft2.deltaT+ float(hof2.epoch))

    plt.plot(tvals2,hoft2.data.data,label='Fr')
    plt.xlim(tcrit-1,tcrit+1)
    plt.plot(tvals,hoft.data.data,label='orig')
    plt.legend(); 

    if not bNoInteractivePlots:
        plt.show()
    else:
        for indx in [1]:
            print("Writing figure ", indx)
            plt.xlim(tcrit-0.1,tcrit+0.01)
            plt.figure(indx); plt.savefig("eob-framedump-" +str(indx)+fig_extension)
#            plt.xlim(min(tvals2),max(tvals2)) # full range with pad
#            plt.figure(indx); plt.savefig("eob-framedump-full-" +str(indx)+fig_extension)





#! /usr/bin/env python
import argparse
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lalframe
import lal

from matplotlib import pyplot as plt

parser = argparse.ArgumentParser()
parser.add_argument("--cache", default="zero_noise.cache")
parser.add_argument("--psd",type=str,default="lalsim.SimNoisePSDaLIGOZeroDetHighPower",help="psd name ('eval'). lalsim.SimNoisePSDaLIGOZeroDetHighPower, lalsim.SimNoisePSDaLIGOZeroDetHighPower, lalsimutils.Wrapper_AdvLIGOPsd, ... ")
parser.add_argument("-p", "--psd-file", action="append", help="instrument=psd-file, e.g. H1=H1_PSD.xml.gz. Can be given multiple times for different instruments.")
parser.add_argument("--fmin-snr",  type=float, default=10)
parser.add_argument("--fmax-snr", type=float, default=1500)
parser.add_argument("--plot-sanity",action='store_true')
opts=  parser.parse_args()

fminSNR = opts.fmin_snr
fmaxSNR = opts.fmax_snr

psd_name = {}
psd_dict = {}
if opts.psd_file:
    for inst, psdf in map(lambda c: c.split("="), opts.psd_file):
        psd_name[inst] = psdf


ifo_list = ["H1", "L1", "V1"]
if opts.psd_file:
    ifo_list = psd_name.keys()

data_dict ={}
rho_dict ={}
rho2Net =0
analyticPSD_Q=True
for ifo in ifo_list:
  try:
    channel = ifo+":FAKE-STRAIN"
    data_dict[ifo] =  lalsimutils.frame_data_to_non_herm_hoff(opts.cache,channel)
    fSample = len(data_dict[ifo].data.data)*data_dict[ifo].deltaF
    df = data_dict[ifo].deltaF
    if not psd_name.has_key(ifo):
        print(ifo, " analytic psd ", opts.psd)
        analyticPSD_Q=True
        psd_dict[ifo] = eval(opts.psd)
    else:
        analyticPSD_Q=False
        print("Reading PSD for instrument %s from %s" % (ifo, psd_name[ifo]))
        psd_dict[ifo] = lalsimutils.load_resample_and_clean_psd(psd_name[ifo], ifo, df)
    IP = lalsimutils.ComplexIP(fLow=fminSNR, fNyq=fSample/2,deltaF=df,psd=psd_dict[ifo],fMax=fmaxSNR,analyticPSD_Q=analyticPSD_Q)
    rhoDet = rho_dict[ifo] = IP.norm(data_dict[ifo])
    print(ifo, rho_dict[ifo])
    rho2Net += rhoDet*rhoDet
    if opts.plot_sanity:
        fvals = lalsimutils.evaluate_fvals(data_dict[ifo])
        plt.plot(fvals, np.power(np.abs(data_dict[ifo].data.data),2) *IP.weights2side)
        plt.xlim(-5000,5000)
        plt.figure(1)
        plt.savefig("frameplot_power_"+ifo+".png"); plt.clf()
        plt.figure(2)
        plt.plot(np.log10(fvals), np.log10(1./IP.weights2side)/2)
        plt. savefig("frameplot_psd_"+ifo+".png"); plt.clf()
    data_dict[ifo] = None  # clear it
  except:
      print(" No IFO ", ifo)
rho_dict['Network']=    np.sqrt(rho2Net )

print(rho_dict)
import json
with open("snr-report.txt", 'w') as f:
    json.dump(rho_dict, f)
    f.flush()




'''
Single Event Neural Network Interpolator
'''

import torch
import os

class Net(torch.nn.Module):
            '''
            Network architecture definition
            '''
        
            def __init__(self, n_inputs, hlayer_size, n_outputs, p_drop):
                super(Net, self).__init__()
                self.linear1 = torch.nn.Linear(n_inputs, hlayer_size)
                self.linear2 = torch.nn.Linear(hlayer_size, hlayer_size)
                self.dropout1 = torch.nn.Dropout(p=p_drop)
                self.linear3 = torch.nn.Linear(hlayer_size, hlayer_size)
                self.dropout2 = torch.nn.Dropout(p=p_drop)
                self.linear4 = torch.nn.Linear(hlayer_size, hlayer_size)
                self.dropout3 = torch.nn.Dropout(p=p_drop)
                self.linear5 = torch.nn.Linear(hlayer_size, hlayer_size)
                self.dropout4 = torch.nn.Dropout(p=p_drop)
                self.linear6 = torch.nn.Linear(hlayer_size, hlayer_size)
                self.dropout5 = torch.nn.Dropout(p=p_drop)
                self.linear7 = torch.nn.Linear(hlayer_size, hlayer_size)
                self.dropout6 = torch.nn.Dropout(p=p_drop)
                self.linear8 = torch.nn.Linear(hlayer_size, hlayer_size)
                self.dropout7 = torch.nn.Dropout(p=p_drop)
                self.linear9 = torch.nn.Linear(hlayer_size, n_outputs)
      
            def forward(self, x):
                x = torch.selu(self.linear1(x))
                x = self.dropout1(torch.selu(self.linear2(x)))
                x = self.dropout2(torch.selu(self.linear3(x)))
                x = self.dropout3(torch.selu(self.linear4(x)))
                x = self.dropout4(torch.selu(self.linear5(x)))
                x = self.dropout5(torch.selu(self.linear6(x)))
                x = self.dropout6(torch.selu(self.linear7(x)))
                x = self.dropout7(torch.selu(self.linear8(x)))
                x = self.linear9(x)
                return x

class Interpolator(object): # interpolator


      def __init__(self, input, target, errors, frac=0.1, test_frac=0.1, hlayer_size=32, p_drop=0, regularize=False,
                   epochs=100, learning_rate=1e-2, betas=(0.9, 0.99), eps=1e-2, weight_decay=1e-6,
                   epochs_per_lr=20, lr_divisions=5, lr_frac=1./3., batch_size=128, shuffle=True, 
                   working_dir='.', loss_func='chi2', no_pad=False):

            '''
            :input: Input column vector with each column representing one input with size (n_samples, n_dim)
            :target: Target column vector of shape (n_samples, 1)
            :frac: The fraction of the input vector taken to be the size of the validation set
            :test_frac: The fraction of the input vector taken to be the size of the test set
            :hlayer_size: Number of hidden neurons in each hidden layer
            :epochs: Number of training epochs
            :learning_rate: Initial learning rate for the optimizer
            :betas: Decay rates for moving averages of gradient (see Adam optimizer documentation)
            :eps: Small number that prevents division by zero in the optimizer
            :epochs_per_lr: Number of epochs that must pass before the learning rate is reduced by lr_frac
            :lr_divisions: Number of times that the learning rate is reduced by lr_frac
            :lr_frac: Fraction by which the learning_rate is multiplied every epochs_per_lr epochs
            :batch_size: Size of the batches used during training
            :shuffle: Boolean deciding whether the data batches are shuffled prior to training
            '''

            import numpy as np
            import os

            try: os.mkdir(working_dir+'/models')
            except: pass

            self.select_device()

            self.working_dir = working_dir
            self.n_inputs = np.size(input, 1)
            self.n_outputs = np.size(target, 1)
            self.hlayer_size = hlayer_size
            self.p_drop = p_drop
            self.regularize=regularize

            self.epochs = epochs
            self.batch_size = batch_size
            self.shuffle = shuffle
            self.store_mu_x =[]
            self.store_sigma_x=[]

            self.store_mu_x = np.zeros(len(input[0]))
            self.store_sigma_x = np.zeros(len(input[0]))

            for dim in xrange(self.n_inputs):
                  self.store_mu_x[dim] = np.mean(input[:,dim])
                  self.store_sigma_x[dim] = np.std(input[:,dim])
                  print( " Computing scaling factors on raw data ", dim, self.store_mu_x[dim], self.store_sigma_x[dim])

            input_train, target_train, errors_train, input_valid, target_valid, errors_valid, input_test, target_test, errors_test \
            = self.set_separation(input, target, errors, frac, test_frac)

            # set the scaling of the 'y' variable to the ORIGINAL problem, not allowing for any padding
            # also use the WHOLE sample to get the scaling factors
            target_copy = np.copy(target)
            _, self.target_mu, self.target_sigma = self.preprocessing(target_copy)
            print( " Computing scaling factors on scaled output ", self.target_mu, self.target_sigma)

            # Create copies to extend, based on input values
            #  - do *not* try to edit these on the fly, since we need to loop through the *original* sample
            input_train_revised = np.array(input_train)
            target_train_revised = np.array(target_train)
            errors_train_revised = np.array(errors_train)
            
            p_epsilon = 1e-3
            
            if not no_pad:
              for dim in xrange(self.n_inputs):
                  # take the max and min values of the array
                  #   - don't use actual min and max, in case of crazy outliers that will leave large gaps in NN coverage
                  true_min,true_max = np.percentile(input_train[:,dim],p_epsilon*100),np.percentile(input_train[:,dim],100*(1-p_epsilon))
                  # find target range
                  #   - choice depends on if signs are different!
                  #   - if same sign, use a factor 
                  test_min, test_max = 0.5*true_min,2*true_max
                  if test_min*test_max < 0: # two different signs for this parameter
                      test_min, test_max = 2*true_min, 2*true_max
                  # Creating fake samples. Note size scaled so TOTAL number of fake samples scales with total sample size
#                  print " Extending range to ", test_min, test_max, input_train[:,dim], "; cutoffs  in physical coordinates are ", self.store_mu_x[dim]+self.store_sigma_x[dim]*np.array([test_min,true_min, true_max, test_max])
                  n_samples_to_add = int(0.3*input_train.shape[0]/(1.0*self.n_inputs))
                  fakesamples = np.random.uniform(test_min, test_max, n_samples_to_add)
                  # remove all the values which occur in the dataset (keep np.where(newv<min) and np.where(newv>max))
                  bad_idxs = np.where((fakesamples > (true_min)) & (fakesamples < (true_max)))
                  fakesamples = np.delete(fakesamples, bad_idxs)
 #                 print " Adding fake samples to dimension ",dim, len(fakesamples), n_samples_to_add, "note mean, width of scaled variable are ",np.mean(input_train[:,dim]),np.std(input_train[:,dim])
                  # set random values (draw from normal) for other parameters and then set likelihood = 0
                  otherdims = np.random.normal(size=(fakesamples.shape[0], self.n_inputs-1))
                  # insert the non-random fakesamples into the relevant dimension index
                  fakesamples = np.insert(otherdims, dim, fakesamples.T, axis=1)
                  zerolnLs = np.ones(fakesamples.shape[0])*1e-5
                  zerolnLs = zerolnLs[:, np.newaxis]
                  unitysigmas = np.ones(fakesamples.shape[0])
                  unitysigmas = unitysigmas[:, np.newaxis]
                  input_train_revised = np.append(input_train_revised, fakesamples, axis=0)
                  target_train_revised = np.append(target_train_revised, zerolnLs, axis=0)
                  errors_train_revised = np.append(errors_train_revised, 0.001*unitysigmas, axis=0) # assume these zeros for lnL are well-known!

            input_train = input_train_revised
            target_train = target_train_revised
            errors_train = errors_train_revised

            # scale the output scale *including* the padded points
            target_train, self.target_mu, self.target_sigma = self.preprocessing(target_train)
            target_valid, _, _ = self.preprocessing(target_valid, self.target_mu, self.target_sigma)
            target_test, _, _ = self.preprocessing(target_test, self.target_mu, self.target_sigma)

            for dim in xrange(self.n_inputs):
                  mu, sigma = self.store_mu_x[dim], self.store_sigma_x[dim]
                  input_train[:, dim], _, _ = self.preprocessing(input_train[:, dim],mu=mu,sigma=sigma)
                  input_valid[:, dim], _, _ = self.preprocessing(input_valid[:, dim], mu=mu, sigma=sigma)
                  input_test[:, dim], _, _ = self.preprocessing(input_test[:, dim], mu=mu, sigma=sigma)

            self.input_train = torch.from_numpy(input_train).float().to(self.device)
            self.input_valid = torch.from_numpy(input_valid).float().to(self.device)
            self.input_test = torch.from_numpy(input_test).float().to(self.device)

            self.target_train = torch.from_numpy(target_train).float().to(self.device)
            self.target_valid = torch.from_numpy(target_valid).float().to(self.device)
            self.target_test = torch.from_numpy(target_test).float().to(self.device)

            self.errors_train = torch.from_numpy(errors_train).float().to(self.device)
            self.errors_valid = torch.from_numpy(errors_valid).float().to(self.device)
            self.errors_test = torch.from_numpy(errors_test).float().to(self.device)

            self.network_init()
            self.loss_func = loss_func
            self.optim_init(learning_rate=learning_rate, betas=betas, eps=eps, weight_decay=weight_decay)
            self.sched_init(epochs_per_lr=epochs_per_lr, lr_divisions=lr_divisions, lr_frac=lr_frac)

            return

      def set_separation(self, input, target, errors, frac, test_frac):
            '''
            Separates the input data array into training, validation, and test sets. Size of validation/test sets is based on fraction of whole input array.
            Both validation and test set have the same size fraction frac unless test_frac is specified.
            '''
            import numpy as np

            idxs = np.random.choice(target.shape[0], int(target.shape[0]*frac), replace=False) # validation set indices

            input_valid, target_valid, errors_valid = input[idxs, :], target[idxs, :], errors[idxs, :]
            input_train, target_train, errors_train = np.delete(input, idxs, 0), np.delete(target, idxs, 0), np.delete(errors, idxs, 0)

            idxs2 = np.random.choice(target_train.shape[0], int(target.shape[0]*test_frac), replace=False) # test set indices

            input_test, target_test, errors_test = input_train[idxs2, :], target_train[idxs2, :], errors_train[idxs2, :]
            input_train, target_train, errors_train = np.delete(input_train, idxs2, 0), np.delete(target_train, idxs2, 0), np.delete(errors_train, idxs2, 0)

            return input_train, target_train, errors_train, input_valid, target_valid, errors_valid, input_test, target_test, errors_test

      def preprocessing(self, data, mu=0, sigma=0):
            '''
            Processes the data such that it is in the proper format (mean-subtracted and sigma-divided) for input into the SELU NN
            If no mu/sigma provided, takes the mean and standard deviation of the data
            '''
            # NOTE: for the validation and test sets, the mean and std of the TRAINING SET are used, not of the respective sets
            import numpy as np

            if not mu and not sigma:
                  data_mu, data_sigma = np.mean(data), np.std(data)
            else:
                  data_mu = mu
                  data_sigma = sigma
            if data_sigma == 0:
                data_sigma = 1  # special case
            data_copy =np.zeros(data.shape)
            data_copy = (data - data_mu)/data_sigma

            return data_copy, data_mu, data_sigma

      def select_device(self):
            '''
            Runs on GPU if available, otherwise runs on CPU
            '''
            if torch.cuda.is_available():
                  device = torch.device('cuda:0')
            else:
                  device = torch.device('cpu')

            self.device = device

      def network_init(self):
            '''
            Network initialization given some number of inputs, outputs, and the size of the hidden layers
            '''
            def weights_init(layer):
                  '''
                  Initialization of weights such that they follow a normal distribution (mu=0, sigma=1)
                  '''
                  import numpy as np
  
                  if isinstance(layer, torch.nn.Linear):
                        torch.nn.init.normal_(layer.weight, std=1./np.sqrt(self.hlayer_size))
            
            if torch.cuda.is_available():
                  self.net = Net(self.n_inputs, self.hlayer_size, self.n_outputs, self.p_drop).cuda().apply(weights_init)
            else:
                  self.net = Net(self.n_inputs, self.hlayer_size, self.n_outputs, self.p_drop).apply(weights_init)

      def optim_init(self, learning_rate, betas, eps, weight_decay):
            '''
            Optimizer initialization, specifically Adam for now, using network parameters and specified optimizer parameters
            '''
            self.optim = torch.optim.Adam(self.net.parameters(), lr=learning_rate, betas=betas, eps=eps, weight_decay=weight_decay)

      def sched_init(self, epochs_per_lr, lr_divisions, lr_frac):
            '''
            Scheduler initialization which selects the epoch intervals at which the learning rate is taken to be frac times the current learning rate.
            Epochs_per_lr indicates how many epochs transpire before the learning rate is divided. 
            Lr_divisions indicates how many times the learning rate is multiplied by frac.
            Frac is a fraction that multiplies the current learning rate to reduce it i.e. initial_lr*(frac^n) after n divisions
            '''
            import numpy as np

            self.epochs_per_lr = epochs_per_lr
            self.lr_divisions = lr_divisions
            self.sched = torch.optim.lr_scheduler.MultiStepLR(self.optim, milestones=np.linspace(self.epochs_per_lr, self.lr_divisions*self.epochs_per_lr, self.lr_divisions), gamma=lr_frac)
      
      def MAPEloss(self, output, target, mu, sigma):
            '''
            Mean absolute percentage error loss function
            '''
            t_max = torch.max(target)
            return 1./output.shape[0]*torch.sum(torch.abs((output-target)*torch.exp(-0.2*torch.abs(t_max-output))/(target+mu/sigma)))

      def reducedchisquareloss(self, output, target, error):

            weight_mag = 0
    
            #for param in self.net.parameters():
            #      param = torch.pow(param, 2)
            #      param = torch.sum(param)
            #      weight_mag += param

            for layer in self.net.children():
                  if not isinstance(layer, torch.nn.Linear): continue
                  weights = layer.weight.data
                  weights = torch.pow(weights, 2)
                  weights = torch.sum(weights)
                  weight_mag += weights
      
            t_max = torch.max(target)
            return torch.sum((output-target)**2*torch.exp(-0.2*torch.abs(t_max-output))/(error)**2)/(output.shape[0]-self.n_inputs) + 1e-6*weight_mag

      def train(self,debug=True):
            '''
            Training on the inputs
            Inputs and target should be of the form [array1, array2, ...]
            '''
            import numpy as np
            from torch.autograd import Variable
            import torch.utils.data as Data

            try: os.system('rm ' + working_dir + '/models/bestmodel.pt')
            except: pass

            dataset = Data.TensorDataset(self.input_train, self.target_train, self.errors_train)

            loader = Data.DataLoader(dataset=dataset, batch_size=self.batch_size, shuffle=self.shuffle)

            validation_threshold = 1e10

            for epoch in np.arange(1, self.epochs+1):

                  for step, (batch_input, batch_target, batch_error) in enumerate(loader):

                        batch_inputs = Variable(batch_input)
                        batch_targets = Variable(batch_target)
                        batch_errors = Variable(batch_error)

                        prediction = self.net(batch_inputs)

                        if self.loss_func == 'mape':
                            loss = self.MAPEloss(prediction, batch_targets, self.target_mu, self.target_sigma)
                        if self.loss_func == 'chi2':
                            loss = self.reducedchisquareloss(prediction, batch_targets, batch_errors)

                        self.optim.zero_grad()
                        loss.backward()      
                        self.optim.step()

                  if self.loss_func == 'mape':
                      validation_loss = self.MAPEloss(self.net(self.input_valid), self.target_valid, self.target_mu, self.target_sigma)
                  if self.loss_func == 'chi2':
                      validation_loss = self.reducedchisquareloss(self.net(self.input_valid), self.target_valid, self.errors_valid)

                  # L2 regularization on *weights*. 
                  # ALREADY implemented in the optimizer with weight_decay: redundant to do both!
                  if self.regularize:
                      reg_loss = None
                      for param in self.net.named_parameters():
                          param_name =param[0]
                          param=param[1]
# Don't regularize last layer (as that sets overall scale of function, rather than structure).
#                      if 'linear9' in param_name:
#                          continue
#  do not regularize bias terms?
#                          if 'bias' in param_name:
#                              continue
                          if reg_loss is None:
                              reg_loss = 0.5 * torch.sum(param**2)
                          else:
                              reg_loss = reg_loss + 0.5 * param.norm(2)**2
                      validation_loss += 0.9*reg_loss
                  

                  if validation_loss < 1e-6:
                      print( "   ... should we stop? ")
                      break

                  if validation_loss < validation_threshold:

                        validation_threshold = validation_loss

                        torch.save({
                        'model_state_dict': self.net.state_dict(),
                        }, self.working_dir+'/models/bestmodel.pt')

                  if epoch % self.epochs_per_lr == 0 and epoch <= self.lr_divisions*self.epochs_per_lr:
                      if os.path.exists(self.working_dir+'/models/bestmodel.pt'):
                        savedmodel = torch.load(self.working_dir+'/models/bestmodel.pt')
                        self.net.load_state_dict(savedmodel['model_state_dict'])

                  self.sched.step()
                  
                  if debug:
                      print( "Epoch %d out of %d complete" % (epoch, self.epochs), '  loss ', validation_loss)

            if self.loss_func == 'mape':
                self.train_loss = self.MAPEloss(self.net(self.input_train), self.target_train, self.target_mu, self.target_sigma)
                self.valid_loss = self.MAPEloss(self.net(self.input_valid), self.target_valid, self.target_mu, self.target_sigma)
                if debug:
                    print( "   Loss (MAPE) ", self.train_loss, self.valid_loss)
            if self.loss_func == 'chi2':
                self.train_loss = self.reducedchisquareloss(self.net(self.input_train), self.target_train, self.errors_train)
                self.valid_loss = self.reducedchisquareloss(self.net(self.input_valid), self.target_valid, self.errors_valid)
                if debug:
                    print( "   Loss (chi2) ", self.train_loss, self.valid_loss)

      def save(self, filename):
            '''
            Saves the model
            '''
            torch.save({
                  'model_state_dict': self.net.state_dict(),
                  'optimizer_state_dict': self.optim.state_dict(),
                  }, filename)
            print( 'Model saved under %s with training loss %f and validation loss %f' % (filename, self.train_loss, self.valid_loss))

      def load(self, filename):
            '''
            Loads a saved model
            '''
            savedmodel = torch.load(filename)
            self.net.load_state_dict(savedmodel['model_state_dict'])
            self.optim.load_state_dict(savedmodel['optimizer_state_dict'])
            print( 'Loaded model %s' % filename)
            return

      def evaluate(self, input):
            '''
            Uses the currently loaded model to evaluate an input array
            '''
            import numpy as np

            input_copy = np.zeros(input.shape)
            for dim in xrange(np.size(input, 1)):
                  input_copy[:, dim], _, _ = self.preprocessing(input[:, dim],mu=self.store_mu_x[dim],sigma=self.store_sigma_x[dim])

            self.net.eval()

            output = self.net(torch.from_numpy(input_copy).float().to(self.device)).detach().cpu().numpy()[:,0] # convert back to 1d output
            output *= self.target_sigma
            output += self.target_mu

            return output

class AdaptiveInterpolator(Interpolator):

      def __init__(self, input, target, errors, frac=0.1, test_frac=0.1, hlayer_size=32, p_drop=0,
                   epochs=100, learning_rate=1e-2, betas=(0.9, 0.99), eps=1e-2, weight_decay=1e-6,
                   epochs_per_lr=20, lr_divisions=5, lr_frac=1./3., batch_size=128, shuffle=True, 
                   working_dir='.', loss_func='chi2', no_pad=False):

            self.input = input
            self.target = target
            self.errors = errors
            self.frac = frac
            self.test_frac = test_frac
            self.hlayer_size = hlayer_size
            self.p_drop=p_drop
            self.epochs = epochs
            self.learning_rate = learning_rate
            self.betas = betas
            self.eps = eps
            self.weight_decay = weight_decay
            self.epochs_per_lr = epochs_per_lr
            self.lr_divisions = lr_divisions
            self.lr_frac = lr_frac
            self.batch_size = batch_size
            self.shuffle = shuffle
            self.working_dir = working_dir
            self.loss_func = loss_func
            self.no_pad = no_pad

            Interpolator.__init__(self, self.input, self.target, self.errors, frac=self.frac, test_frac=self.test_frac, hlayer_size=self.hlayer_size, p_drop=self.p_drop,
                   epochs=self.epochs, learning_rate=self.learning_rate, betas=self.betas, eps=self.eps, weight_decay=self.weight_decay, epochs_per_lr=self.epochs_per_lr, 
                   lr_divisions=self.lr_divisions, lr_frac=self.lr_frac, batch_size=self.batch_size, shuffle=self.shuffle, working_dir=self.working_dir,
                   loss_func=self.loss_func, no_pad=self.no_pad)

      def train(self):

            import numpy as np

            print('Using adaptive training...')

            super(AdaptiveInterpolator, self).train()

            while self.valid_loss < 1e-3:

                  self.hlayer_size /= 2

                  print('Decreasing until reasonable layer size reached, currently trying %d' % self.hlayer_size)

                  Interpolator.__init__(self, self.input, self.target, self.errors, frac=self.frac, test_frac=self.test_frac, hlayer_size=self.hlayer_size, p_drop=self.p_drop,
                   epochs=self.epochs, learning_rate=self.learning_rate, betas=self.betas, eps=self.eps, weight_decay=self.weight_decay, epochs_per_lr=self.epochs_per_lr, 
                   lr_divisions=self.lr_divisions, lr_frac=self.lr_frac, batch_size=self.batch_size, shuffle=self.shuffle, working_dir=self.working_dir,
                   loss_func=self.loss_func, no_pad=self.no_pad)

                  super(AdaptiveInterpolator, self).train()

            if self.valid_loss < 1: self.hlayer_size /= 2

            ceiling = 2*self.hlayer_size

            floor = self.hlayer_size

            layer_sizes = np.linspace(floor, ceiling, 5)

            losses = np.zeros_like(layer_sizes)

            for size in layer_sizes:

                  self.hlayer_size = int(size)

                  print('Looping through layer sizes, currently at %d' % self.hlayer_size)

                  Interpolator.__init__(self, self.input, self.target, self.errors, frac=self.frac, test_frac=self.test_frac, hlayer_size=self.hlayer_size, p_drop=self.p_drop,
                   epochs=self.epochs, learning_rate=self.learning_rate, betas=self.betas, eps=self.eps, weight_decay=self.weight_decay, epochs_per_lr=self.epochs_per_lr, 
                   lr_divisions=self.lr_divisions, lr_frac=self.lr_frac, batch_size=self.batch_size, shuffle=self.shuffle, working_dir=self.working_dir,
                   loss_func=self.loss_func, no_pad=self.no_pad)

                  super(AdaptiveInterpolator, self).train()

                  losses[np.where(layer_sizes==size)] = np.abs(self.valid_loss.detach().cpu().numpy() - 1)

            self.hlayer_size = int(layer_sizes[np.argmin(losses)])

            self.losses = losses
            self.layer_sizes = layer_sizes

            Interpolator.__init__(self, self.input, self.target, self.errors, frac=self.frac, test_frac=self.test_frac, hlayer_size=self.hlayer_size, p_drop=self.p_drop,
                   epochs=self.epochs, learning_rate=self.learning_rate, betas=self.betas, eps=self.eps, weight_decay=self.weight_decay, epochs_per_lr=self.epochs_per_lr, 
                   lr_divisions=self.lr_divisions, lr_frac=self.lr_frac, batch_size=self.batch_size, shuffle=self.shuffle, working_dir=self.working_dir,
                   loss_func=self.loss_func, no_pad=self.no_pad)

            print( '-------OPTIMAL HIDDEN LAYERS FOUND TO BE %d, RE-TRAINING WITH THIS SETTING-------' % self.hlayer_size)

            super(AdaptiveInterpolator, self).train()



#! /usr/bin/env python
#
# GOAL
#   - use downselect-parameter, downselect-parameter-range to filter *.xml.gz files for target samples
#
#


import argparse
import sys
import numpy as np
import RIFT.lalsimutils as lalsimutils
import lalframe
import lal
import functools
import itertools


parser = argparse.ArgumentParser()
parser.add_argument("--fname",default=None)
parser.add_argument("--inj",default=None,help="Reference configuration")
parser.add_argument("--event",default=0,type=int,help="Reference event in the inj file")
parser.add_argument("--fname-out",default="filtered-output.xml.gz")
# Parameters
parser.add_argument("--downselect-parameter",action='append', help='Name of parameter to be used to eliminate grid points ')
parser.add_argument("--downselect-parameter-range",action='append',type=str)
parser.add_argument("--verbose",action='store_true')
opts=  parser.parse_args()

downselect_dict = {}
if opts.downselect_parameter:
    dlist = opts.downselect_parameter
    dlist_ranges  = map(eval,opts.downselect_parameter_range)
else:
    dlist = []
    dlist_ranges = []
if len(dlist) != len(dlist_ranges):
    print(" downselect parameters inconsistent", dlist, dlist_ranges)
for indx in np.arange(len(dlist_ranges)):
    downselect_dict[dlist[indx]] = dlist_ranges[indx]

if otps.verbose:
    print(downselect_dict)

# downselection procedure: fix units so I can downselect on mass parameters
for p in ['mc', 'm1', 'm2', 'mtot']:
    if p in downselect_dict.keys():
        downselect_dict[p] = np.array(downselect_dict[p],dtype=np.float64)
        downselect_dict[p] *= lal.MSUN_SI  # add back units


P_list_in = lalsimutils.xml_to_ChooseWaveformParams_array(opts.fname)
P_inj = lalsimutils.xml_to_ChooseWaveformParams_array(opts.fname)[opts.event]
param_ref_vals = {}
for param in downselect_dict.keys():
    param_ref_vals[param] = P_inj.extract_param(param)

P_list = []
for P in P_list_in:
    include_item = True
    for param in downselect_dict.keys():
        param_here = P.extract_param(param)
        if not (param_here > downselect_dict[param][0] and param_here < downselect_dict[param][1]):
            include_item=False
            continue
    if include_item:
        P_list.append(P)


lalsimutils.ChooseWaveformParams_array_to_xml(P_list, opts.fname_out.replace(".xml.gz", "")) # don't write the postfix twice



#! /usr/bin/env python
#
# GOAL
#   - load in lnL data
#   - fit peak to quadratic (standard), GP, etc. 
#   - evaluate,  based on some parameter grid
#
# FORMAT
#   - pankow simplification of standard format
#
# COMPARE TO
#   util_NRQuadraticFit.py
#   postprocess_1d_cumulative
#   util_QuadraticMassPosterior.py
#


import RIFT.interpolators.BayesianLeastSquares as BayesianLeastSquares

import argparse
import sys
import numpy as np
import numpy.lib.recfunctions
import scipy
import RIFT.lalsimutils as lalsimutils
import lalsimulation as lalsim
import lalframe
import lal
import functools
import itertools

no_plots = True


try:
    import matplotlib.pyplot as plt
    from mpl_toolkits.mplot3d import Axes3D
    import matplotlib.lines as mlines
    import corner

    no_plots=False
except ImportError:
    print(" - no matplotlib - ")


from sklearn.preprocessing import PolynomialFeatures
import RIFT.misc.ModifiedScikitFit as msf  # altenative polynomialFeatures
from sklearn import linear_model

from glue.ligolw import lsctables, utils, ligolw
lsctables.use_in(ligolw.LIGOLWContentHandler)

import RIFT.integrators.mcsampler as mcsampler


def render_coord(x):
    if x in lalsimutils.tex_dictionary.keys():
        return tex_dictionary[x]
    if 'product(' in x:
        a=x.replace(' ', '') # drop spaces
        a = a[:len(a)-1] # drop last
        a = a[8:]
        terms = a.split(',')
        exprs =map(render_coord, terms)
        exprs = map( lambda x: x.replace('$', ''), exprs)
        my_label = ' '.join(exprs)
        return '$'+my_label+'$'
    else:
        return x

def render_coordinates(coord_names):
    return map(render_coord, coord_names)


parser = argparse.ArgumentParser()
parser.add_argument("--maximize-mass",action='store_true', help="If true, maximize the likelihood for each value of the total mass. Ignore any grid placement in total mass")
parser.add_argument("--fname",help="filename of *.dat file [standard ILE output]")
parser.add_argument("--input-tides",action='store_true',help="Use input format with tidal fields included.")
parser.add_argument("--fname-xml-base",help="filename of xml file to use as base (e.g., to specify m1,m2, chi1, chi2, ... to minimize burden on ascii file")
parser.add_argument("--fname-parameter-grid",help="filename of ascii parameters to use to evaluate the fit")
parser.add_argument("--fname-out",default="eval.dat")
parser.add_argument("--fref",default=20,type=float, help="Reference frequency used for spins in the ILE output.  (Since I usually use SEOBNRv3, the best choice is 20Hz)")
parser.add_argument("--fmin",type=float,default=20)
parser.add_argument("--fname-rom-samples",default=None,help="*.rom_composite output. Treated identically to set of posterior samples produced by mcsampler after constructing fit.")
parser.add_argument("--parameter", action='append', help="Parameters used as fitting parameters AND varied at a low level to make a posterior")
parser.add_argument("--parameter-implied", action='append', help="Parameter used in fit, but not independently varied for Monte Carlo")
parser.add_argument("--mc-range",default=None,help="Chirp mass range [mc1,mc2]. Important if we have a low-mass object, to avoid wasting time sampling elsewhere.")
parser.add_argument("--eta-range",default=None,help="Eta range. Important if we have a BNS or other item that has a strong constraint.")
parser.add_argument("--mtot-range",default=None,help="Chirp mass range [mc1,mc2]. Important if we have a low-mass object, to avoid wasting time sampling elsewhere.")
parser.add_argument("--trust-sample-parameter-box",action='store_true', help="If used, sets the prior range to the SAMPLE range for any parameters. NOT IMPLEMENTED. This should be automatically done for mc!")
parser.add_argument("--plots-do-not-force-large-range",action='store_true', help = "If used, the plots do NOT automatically set the chieff range to [-1,1], the eta range to [0,1/4], etc")
parser.add_argument("--downselect-parameter",action='append', help='Name of parameter to be used to eliminate grid points ')
parser.add_argument("--downselect-parameter-range",action='append',type=str)
parser.add_argument("--chi-max", default=1,type=float,help="Maximum range of 'a' allowed.  Use when comparing to models that aren't calibrated to go to the Kerr limit.")
parser.add_argument("--parameter-nofit", action='append', help="Parameter used to initialize the implied parameters, and varied at a low level, but NOT the fitting parameters")
parser.add_argument("--use-precessing",action='store_true')
parser.add_argument("--lnL-offset",type=float,default=10,help="lnL offset")
parser.add_argument("--lnL-cut",type=float,default=None,help="lnL cut [MANUAL]")
parser.add_argument("--M-max-cut",type=float,default=1e5,help="Maximum mass to consider (e.g., if there is a cut on distance, this matters)")
parser.add_argument("--sigma-cut",type=float,default=0.6,help="Eliminate points with large error from the fit.")
parser.add_argument("--ignore-errors-in-data",action='store_true',help='Ignore reported error in lnL. Helpful for testing purposes (i.e., if the error is zero)')
parser.add_argument("--lnL-peak-insane-cut",type=float,default=np.inf,help="Throw away lnL greater than this value. Should not be necessary")
parser.add_argument("--verbose", action="store_true",default=False, help="Required to build post-frame-generating sanity-test plots")
parser.add_argument("--save-plots",default=False,action='store_true', help="Write plots to file (only useful for OSX, where interactive is default")
parser.add_argument("--inj-file", help="Name of injection file")
parser.add_argument("--event-num", type=int, default=0,help="Zero index of event in inj_file")
parser.add_argument("--report-best-point",action='store_true')
parser.add_argument("--adapt",action='store_true')
parser.add_argument("--fit-uses-reported-error",action='store_true')
parser.add_argument("--fit-uses-reported-error-factor",type=float,default=1,help="Factor to add to standard deviation of fit, before adding to lnL. Multiplies number fitting dimensions")
parser.add_argument("--n-max",default=3e5,type=float)
parser.add_argument("--n-eff",default=3e3,type=int)
parser.add_argument("--fit-method",default="quadratic",help="quadratic|polynomial|gp|gp_hyper")
parser.add_argument("--fit-order",type=int,default=2,help="Fit order (polynomial case: degree)")
parser.add_argument("--fit-uncertainty-added",default=False, action='store_true', help="Reported likelihood is lnL+(fit error). Use for placement and use of systematic errors.")
parser.add_argument("--no-plots",action='store_true')
opts=  parser.parse_args()
no_plots = no_plots |  opts.no_plots

with open('args.txt','w') as fp:
    import sys
    fp.write(' '.join(sys.argv))

if opts.fit_method == "quadratic":
    opts.fit_order = 2  # overrride

###
### Comparison data (from LI)
###
remap_ILE_2_LI = {
 "s1z":"a1z", "s2z":"a2z", 
 "s1x":"a1x", "s1y":"a1y",
 "s2x":"a2x", "s2y":"a2y",
 "chi1_perp":"chi1_perp",
 "chi2_perp":"chi2_perp",
 "chi1":'a1',
 "chi2":'a2',
 "cos_phiJL": 'cos_phiJL',
 "sin_phiJL": 'sin_phiJL',
 "cos_theta1":'costilt1',
 "cos_theta2":'costilt2',
 "theta1":"tilt1",
 "theta2":"tilt2",
  "xi":"chi_eff", 
  "chiMinus":"chi_minus", 
  "delta":"delta", 
 "mtot":'mtotal', "mc":"mc", "eta":"eta","m1":"m1","m2":"m2",
  "cos_beta":"cosbeta",
  "beta":"beta",
  "LambdaTilde":"lambdat",
  "DeltaLambdaTilde": "dlambdat"}

downselect_dict = {}
dlist = []
dlist_ranges=[]
if opts.downselect_parameter:
    dlist = opts.downselect_parameter
    dlist_ranges  = map(eval,opts.downselect_parameter_range)
else:
    dlist = []
    dlist_ranges = []
if len(dlist) != len(dlist_ranges):
    print(" downselect parameters inconsistent", dlist, dlist_ranges)
for indx in np.arange(len(dlist_ranges)):
    downselect_dict[dlist[indx]] = dlist_ranges[indx]


chi_max = opts.chi_max
downselect_dict['chi1'] = [0,chi_max]
downselect_dict['chi2'] = [0,chi_max]
for param in ['s1z', 's2z', 's1x','s2x', 's1y', 's2y']:
    downselect_dict[param] = [-chi_max,chi_max]
# Enforce definition of eta
downselect_dict['eta'] = [0,0.25]



test_converged={}
#test_converged['neff'] = functools.partial(mcsampler.convergence_test_MostSignificantPoint,0.01)  # most significant point less than 1/neff of probability.  Exactly equivalent to usual neff threshold.
#test_converged["normal_integral"] = functools.partial(mcsampler.convergence_test_NormalSubIntegrals, 25, 0.01, 0.1)   # 20 sub-integrals are gaussian distributed [weakly; mainly to rule out outliers] *and* relative error < 10%, based on sub-integrals . Should use # of intervals << neff target from above.  Note this sets our target error tolerance on  lnLmarg.  Note the specific test requires >= 20 sub-intervals, which demands *very many* samples (each subintegral needs to be converged).


prior_range_map = {"mtot": [1, 300], "q":[0.01,1], "s1z":[-0.999*chi_max,0.999*chi_max], "s2z":[-0.999*chi_max,0.999*chi_max], "mc":[0.9,250], "eta":[0.01,0.2499999], 'xi':[-chi_max,chi_max],'chi_eff':[-chi_max,chi_max],'delta':[-1,1],
   's1x':[-chi_max,chi_max],
   's2x':[-chi_max,chi_max],
   's1y':[-chi_max,chi_max],
   's2y':[-chi_max,chi_max],
  'm1':[0.9,1e3],
  'm2':[0.9,1e3],
  'lambda1':[0.01,4000],
  'lambda2':[0.01,4000],
  # strongly recommend you do NOT use these as parameters!  Only to insure backward compatibility with LI results
  'LambdaTilde':[0.01,5000],
  'DeltaLambdaTilde':[-500,500],
}

if not (opts.eta_range is None):
    print(" Warning: Overriding default eta range. USE WITH CARE")
    prior_range_map['eta'] = eval(opts.eta_range)  # really only useful if eta is a coordinate.  USE WITH CARE




# TeX dictionary
tex_dictionary = lalsimutils.tex_dictionary


###
### Linear fits. Resampling a quadratic. (Export me)
###

def fit_quadratic_alt(x,y,y_err=None,x0=None,symmetry_list=None,verbose=False):
    gamma_x = None
    if not (y_err is None):
        gamma_x =1./np.power(y_err,2)
    the_quadratic_results = BayesianLeastSquares.fit_quadratic( x, y,gamma_x=gamma_x,verbose=verbose)#x0=None)#x0_val_here)
    peak_val_est, best_val_est, my_fisher_est, linear_term_est,fn_estimate = the_quadratic_results

    np.savetxt("lnL_peakval.dat",[peak_val_est])   # generally not very useful
    np.savetxt("lnL_bestpt.dat",best_val_est)  
    np.savetxt("lnL_gamma.dat",my_fisher_est,header=' '.join(coord_names))
        

    bic  =-2*( -0.5*np.sum(np.power((y - fn_estimate(x)),2))/2 - 0.5* len(y)*np.log(len(x[0])) )

    print("  Fit: std :" , np.std( y-fn_estimate(x)))
    print("  Fit: BIC :" , bic)

    return fn_estimate


# https://github.com/scikit-learn/scikit-learn/blob/14031f6/sklearn/preprocessing/data.py#L1139
def fit_polynomial(x,y,x0=None,symmetry_list=None,y_errors=None):
    """
    x = array so x[0] , x[1], x[2] are points.
    """

    clf_list = []
    bic_list = []
    for indx in np.arange(opts.fit_order+1):
        poly = msf.PolynomialFeatures(degree=indx,symmetry_list=symmetry_list)
        X_  = poly.fit_transform(x)

        if opts.verbose:
            print(" Fit : poly: RAW :", poly.get_feature_names())
            print(" Fit : ", poly.powers_)

        # Strip things with inappropriate symmetry: IMPOSSIBLE
        # powers_new = []
        # if not(symmetry_list is None):
        #  for line in poly.powers_:
        #     signature = np.prod(np.power( np.array(symmetry_list), line))
        #     if signature >0:
        #         powers_new.append(line)
        #  poly.powers_ = powers_new

        #  X_  = poly.fit_transform(x) # refit, with symmetry-constrained structure

        #  print " Fit : poly: After symmetry constraint :", poly.get_feature_names()
        #  print " Fit : ", poly.powers_


        clf = linear_model.LinearRegression()
        if y_errors is None or opts.ignore_errors_in_data:
            clf.fit(X_,y)
        else:
            assert len(y_errors) == len(y)
            clf.fit(X_,y,sample_weight=1./y_errors**2)  # fit with usual weights

        clf_list.append(clf)

        print(" Fit: Testing order ", indx)
        print(" Fit: std: ", np.std(y - clf.predict(X_)),  "using number of features ", len(y))  # should NOT be perfect
        if not (y_errors is None):
            print(" Fit: weighted error ", np.std( (y - clf.predict(X_))/y_errors))
        bic = -2*( -0.5*np.sum(np.power(y - clf.predict(X_),2))  - 0.5*len(y)*np.log(len(x[0])))
        print(" Fit: BIC:", bic)
        bic_list.append(bic)

    clf = clf_list[np.argmin(np.array(bic_list) )]

    return lambda x: clf.predict(poly.fit_transform(x))


from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, WhiteKernel, ConstantKernel as C

def adderr(y):
    val,err = y
    return val+error_factor*err

def fit_gp(x,y,x0=None,symmetry_list=None,y_errors=None,hypercube_rescale=False):
    """
    x = array so x[0] , x[1], x[2] are points.
    """

    # Amplitude: 
    #   - We are fitting lnL.  
    #   - We know the scale more or less: more than 2 in the log is bad
    # Scale
    #   - because of strong correlations with chirp mass, the length scales can be very short
    #   - they are rarely very long, but at high mass can be long
    #   - I need to allow for a RANGE

    length_scale_est = []
    length_scale_bounds_est = []
    for indx in np.arange(len(x[0])):
        # These length scales have been tuned by expereience
        length_scale_est.append( 2*np.std(x[:,indx])  )  # auto-select range based on sampling retained
        length_scale_min_here= np.max([1e-3,0.2*np.std(x[:,indx]/np.sqrt(len(x)))])
        if indx == mc_index:
            length_scale_min_here= 0.2*np.std(x[:,indx]/np.sqrt(len(x)))
            print(" Setting mc range: retained point range is ", np.std(x[:,indx]), " and target min is ", length_scale_min_here)
        length_scale_bounds_est.append( (length_scale_min_here , 5*np.std(x[:,indx])   ) )  # auto-select range based on sampling *RETAINED* (i.e., passing cut).  Note that for the coordinates I usually use, it would be nonsensical to make the range in coordinate too small, as can occasionally happens

    print(" GP: Estimated length scales ")
    print(length_scale_est)
    print(length_scale_bounds_est)

    if not (hypercube_rescale):
        # These parameters have been hand-tuned by experience to try to set to levels comparable to typical lnL Monte Carlo error
        kernel = WhiteKernel(noise_level=0.1,noise_level_bounds=(1e-2,1))+C(0.5, (1e-3,1e1))*RBF(length_scale=length_scale_est, length_scale_bounds=length_scale_bounds_est)
        gp = GaussianProcessRegressor(kernel=kernel, n_restarts_optimizer=8)

        gp.fit(x,y)

        print(" Fit: std: ", np.std(y - gp.predict(x)),  "using number of features ", len(y))

        if not (opts.fit_uncertainty_added):
            return lambda x: gp.predict(x)
        else:
            return lambda x: adderr(gp.predict(x,return_std=True))
    else:
        x_scaled = np.zeros(x.shape)
        x_center = np.zeros(len(length_scale_est))
        x_center = np.mean(x)
        print(" Scaling data to central point ", x_center)
        for indx in np.arange(len(x)):
            x_scaled[indx] = (x[indx] - x_center)/length_scale_est # resize

        kernel = WhiteKernel(noise_level=0.1,noise_level_bounds=(1e-2,1))+C(0.5, (1e-3,1e1))*RBF( len(x_center), (1e-3,1e1))
        gp = GaussianProcessRegressor(kernel=kernel, n_restarts_optimizer=8)
        
        gp.fit(x_scaled,y)
        print(" Fit: std: ", np.std(y - gp.predict(x_scaled)),  "using number of features ", len(y))  # should NOT be perfect

        return lambda x,x0=x_center,scl=length_scale_est: gp.predict( (x-x0 )/scl)

coord_names = opts.parameter # Used  in fit
if coord_names is None:
    coord_names = []
low_level_coord_names = coord_names # Used for Monte Carlo
if opts.parameter_implied:
    coord_names = coord_names+opts.parameter_implied
if opts.parameter_nofit:
    if opts.parameter is None:
        low_level_coord_names = opts.parameter_nofit # Used for Monte Carlo
    else:
        low_level_coord_names = opts.parameter+opts.parameter_nofit # Used for Monte Carlo
error_factor = len(coord_names)
if opts.fit_uses_reported_error:
    error_factor=len(coord_names)*opts.fit_uses_reported_error_factor
print(" Coordinate names for fit :, ", coord_names)
print(" Rendering coordinate names : ",  render_coordinates(coord_names))  # map(lambda x: tex_dictionary[x], coord_names)
print(" Symmetry for these fitting coordinates :", lalsimutils.symmetry_sign_exchange(coord_names))
print(" Coordinate names for Monte Carlo :, ", low_level_coord_names)
print(" Rendering coordinate names : ", map(lambda x: tex_dictionary[x], low_level_coord_names))

# initialize
dat_mass  = [] 
weights = []
n_params = -1

###
### Retrieve data
###
#  id m1 m2  lnL sigma/L  neff
col_lnL = 9
if opts.input_tides:
    print(" Tides input")
    col_lnL +=2
dat_orig = dat = np.loadtxt(opts.fname)
dat_orig = dat[dat[:,col_lnL].argsort()] # sort  http://stackoverflow.com/questions/2828059/sorting-arrays-in-numpy-by-column
print(" Original data size = ", len(dat), dat.shape)

 ###
 ### Convert data.  Use lalsimutils for flexibility
 ###
P_list = []
dat_out =[]
 

symmetry_list =lalsimutils.symmetry_sign_exchange(coord_names)  # identify symmetry due to exchange
mc_min = 1e10
mc_max = -1

mc_index = -1 # index of mchirp in parameter index. To help with nonstandard GP
mc_cut_range = [-np.inf, np.inf] 
if opts.mc_range:
    mc_cut_range = eval(opts.mc_range)  # throw out samples outside this range
print(" Stripping samples outside of ", mc_cut_range, " in mc")
P= lalsimutils.ChooseWaveformParams()
for line in dat:
  # Skip precessing binaries unless explicitly requested not to!
  if not opts.use_precessing and (line[3]**2 + line[4]**2 + line[6]**2 + line[7]**2)>0.01:
      print(" Skipping precessing binaries ")
      continue
  if line[1]+line[2] > opts.M_max_cut:
      if opts.verbose:
          print(" Skipping ", line, " as too massive, with mass ", line[1]+line[2])
      continue
  if line[col_lnL+1] > opts.sigma_cut:
#      if opts.verbose:
#          print " Skipping ", line
      continue
  if line[col_lnL] < opts.lnL_cut:
      continue  # strip worthless points.  DANGEROUS
  mc_here = lalsimutils.mchirp(line[1],line[2])
  if mc_here < mc_cut_range[0] or mc_here > mc_cut_range[1]:
      if opts.verbose:
          print("Stripping because sample outside of target  mc range ", line)
      continue
  if line[col_lnL] < opts.lnL_peak_insane_cut:
    P.fref = opts.fref  # IMPORTANT if you are using a quantity that depends on J
    P.fmin = opts.fmin
    P.m1 = line[1]*lal.MSUN_SI
    P.m2 = line[2]*lal.MSUN_SI
    P.s1x = line[3]
    P.s1y = line[4]
    P.s1z = line[5]
    P.s2x = line[6]
    P.s2y = line[7]
    P.s2z = line[8]

    if opts.input_tides:
        P.lambda1 = line[9]
        P.lambda2 = line[10]

    # INPUT GRID: Evaluate binary parameters on fitting coordinates
    line_out = np.zeros(len(coord_names)+2)
    for x in np.arange(len(coord_names)):
        line_out[x] = P.extract_param(coord_names[x])
 #        line_out[x] = getattr(P, coord_names[x])
    line_out[-2] = line[col_lnL]
    line_out[-1] = line[col_lnL+1]  # adjoin error estimate
    dat_out.append(line_out)


    # results using sampling coordinates (low_level_coord_names) 
    line_out = np.zeros(len(low_level_coord_names))
    for x in np.arange(len(line_out)):
        fac = 1
        if low_level_coord_names[x] in ['mc','m1','m2','mtot']:
            fac = lal.MSUN_SI
        line_out[x] = P.extract_param(low_level_coord_names[x])/fac
        if low_level_coord_names[x] in ['mc','mtot']:  # only use one overall mass index
            mc_index = x


    # Update mc range
    mc_here = lalsimutils.mchirp(line[1],line[2])
    if mc_here < mc_min:
        mc_min = mc_here
    if mc_here > mc_max:
        mc_max = mc_here

Pref_default = P.copy()  # keep this around to fix the masses, if we don't have an inj

dat_out = np.array(dat_out)
print(" Stripped size  = ", dat_out.shape)
 # scale out mass units
for p in ['mc', 'm1', 'm2', 'mtot']:
    if p in coord_names:
        indx = coord_names.index(p)
        dat_out[:,indx] /= lal.MSUN_SI
            


# Repack data
X =dat_out[:,0:len(coord_names)]
Y = dat_out[:,-2]
Y_err = dat_out[:,-1]

# Eliminate values with Y too small
max_lnL = np.max(Y)
indx_ok = Y>np.max(Y)-opts.lnL_offset
print(" Points used in fit : ", sum(indx_ok), " given max lnL ", max_lnL)
if max_lnL < 10:
    # nothing matters, we will reject it anyways
    indx_ok = np.ones(len(Y),dtype=bool)
elif sum(indx_ok) < 10: # and max_lnL > 30:
    # mark the top 10 elements and use them for fits
    # this may be VERY VERY DANGEROUS if the peak is high and poorly sampled
    idx_sorted_index = np.lexsort((np.arange(len(Y)), Y))  # Sort the array of Y, recovering index values
    indx_list = np.array( [[k, Y[k]] for k in idx_sorted_index])     # pair up with the weights again
    indx_list = indx_list[::-1]  # reverse, so most significant are first
    indx_ok = map(int,indx_list[:10,0])
    print(" Revised number of points for fit: ", sum(indx_ok), indx_ok, indx_list[:10])
X_raw = X.copy()

my_fit= None
if opts.fit_method == "quadratic":
    print(" FIT METHOD ", opts.fit_method, " IS QUADRATIC")
    X=X[indx_ok]
    Y=Y[indx_ok]
    Y_err = Y_err[indx_ok]
    my_fit = fit_quadratic_alt(X,Y,symmetry_list=symmetry_list,verbose=opts.verbose)
elif opts.fit_method == "polynomial":
    print(" FIT METHOD ", opts.fit_method, " IS POLYNOMIAL")
    X=X[indx_ok]
    Y=Y[indx_ok]
    Y_err = Y_err[indx_ok]
    my_fit = fit_polynomial(X,Y,symmetry_list=symmetry_list,y_errors=Y_err)
elif opts.fit_method == 'gp_hyper':
    print(" FIT METHOD ", opts.fit_method, " IS GP with hypercube rescaling")
    # some data truncation IS used for the GP, but beware
    print(" Truncating data set used for GP, to reduce memory usage needed in matrix operations")
    X=X[indx_ok]
    Y=Y[indx_ok]
    Y_err = Y_err[indx_ok]
    my_fit = fit_gp(X,Y,y_errors=Y_err,hypercube_rescale=True)
elif opts.fit_method == 'gp':
    print(" FIT METHOD ", opts.fit_method, " IS GP")
    # some data truncation IS used for the GP, but beware
    print(" Truncating data set used for GP, to reduce memory usage needed in matrix operations")
    X=X[indx_ok]
    Y=Y[indx_ok]
    Y_err = Y_err[indx_ok]
    my_fit = fit_gp(X,Y,y_errors=Y_err)

# Sort for later convenience (scatterplots, etc)
indx = Y.argsort()#[::-1]
X=X[indx]
Y=Y[indx]






###
### Coordinate conversion tool
###
def convert_coords(x_in):
    return lalsimutils.convert_waveform_coordinates(x_in, coord_names=coord_names,low_level_coord_names=low_level_coord_names)



likelihood_function = None
if len(low_level_coord_names) ==1:
    def likelihood_function(x):  
        if isinstance(x,float):
            return np.exp(my_fit([x]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x]).T)  ))
if len(low_level_coord_names) ==2:
    def likelihood_function(x,y):  
        if isinstance(x,float):
            return np.exp(my_fit([x,y]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x,y]).T)))
if len(low_level_coord_names) ==3:
    def likelihood_function(x,y,z):  
        if isinstance(x,float):
            return np.exp(my_fit([x,y,z]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x,y,z]).T)))
if len(low_level_coord_names) ==4:
    def likelihood_function(x,y,z,a):  
        if isinstance(x,float):
            return np.exp(my_fit([x,y,z,a]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x,y,z,a]).T)))
if len(low_level_coord_names) ==5:
    def likelihood_function(x,y,z,a,b):  
        if isinstance(x,float):
            return np.exp(my_fit([x,y,z,a,b]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x,y,z,a,b]).T)))
if len(low_level_coord_names) ==6:
    def likelihood_function(x,y,z,a,b,c):  
        if isinstance(x,float):
            return np.exp(my_fit([x,y,z,a,b,c]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x,y,z,a,b,c]).T)))
if len(low_level_coord_names) ==7:
    def likelihood_function(x,y,z,a,b,c,d):  
        if isinstance(x,float):
            return np.exp(my_fit([x,y,z,a,b,c,d]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x,y,z,a,b,c,d]).T)))
if len(low_level_coord_names) ==8:
    def likelihood_function(x,y,z,a,b,c,d,e):  
        if isinstance(x,float):
            return np.exp(my_fit([x,y,z,a,b,c,d,e]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x,y,z,a,b,c,d,e]).T)))
if len(low_level_coord_names) ==9:
    def likelihood_function(x,y,z,a,b,c,d,e,f):  
        if isinstance(x,float):
            return np.exp(my_fit([x,y,z,a,b,c,d,e,f]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x,y,z,a,b,c,d,e,f]).T)))
if len(low_level_coord_names) ==10:
    def likelihood_function(x,y,z,a,b,c,d,e,f,g):  
        if isinstance(x,float):
            return np.exp(my_fit([x,y,z,a,b,c,d,e,f,g]))
        else:
            return np.exp(my_fit(convert_coords(np.array([x,y,z,a,b,c,d,e,f,g]).T)))




# PROCEDURE
#   - Identify grid of desired parameters (e.g., ascii table) 
#   - Create

# Base
P_base = lalsimutils.xml_to_ChooseWaveformParams_array(opts.fname_xml_base)[0]

# Grid
samples_rec = np.genfromtxt(opts.fname_parameter_grid,names=True)
params_rec = samples_rec.dtype.names

# Conversion
P_list  =[]
grid_list = []
lnL_list = []
for indx in np.arange(len(samples_rec[params_rec[0]])):
    P = P_base.manual_copy()
    for param in params_rec:
        val = samples_rec[param][indx]
        fac=1
        if param in ['mc','m1','m2','mtot']:
            fac = lal.MSUN_SI
        P.assign_param(param,fac*val)

    if opts.verbose:
        P.print_params()
    line_out = np.zeros(len(coord_names))
    for x in np.arange(len(line_out)):
        fac = 1
        if coord_names[x] in ['mc','m1','m2','mtot']:
            fac = lal.MSUN_SI
        line_out[x] = P.extract_param(coord_names[x])/fac

    # If opts.maximize_mass, we are reporting the likelihood maximized in total mass (all other parameters held fixed)
    # Remember, mc_index tells us the variable we need to scale
    arg=-1
    if (not opts.maximize_mass) or mc_index <0:
        arg = my_fit(line_out)[0]
    else:
        scalevec = np.ones(len(coord_names));
        def scaledfunc(x):
            scalevec[mc_index] = x
            val = -my_fit(line_out*scalevec)
            return -my_fit(line_out*scalevec)[0]
        res= scipy.optimize.minimize(scaledfunc,1,bounds=[(0.01,100)],options={'maxiter':50})  # unlikely to have mass range scale of a factor of 10^4
        arg = -scaledfunc(res.x)
    grid_list.append(line_out)
    lnL_list.append(arg)
    print(line_out, arg)


n_params = len(grid_list[0])
dat_out = np.zeros( (len(grid_list), n_params+1))
dat_out[:,:n_params] = np.array(grid_list)
dat_out[:,-1] = np.array(lnL_list)
np.savetxt(opts.fname_out, dat_out)
